from loguru import logger

from sweepai.backend.api_utils import (
    get_pr_snippets,
)
from sweepai.core.entities import (
    AssistantMessage,
    Message,
    Snippet,
    SystemMessage,
    UserMessage,
)
from sweepai.core.github_utils import ClonedRepo
from sweepai.core.llm.chat import LATEST_CLAUDE_SONNET_MODEL, Thread, continuous_llm_calls
from sweepai.dataclasses.chat_agent_context import ChatAgentContext
from sweepai.dataclasses.code_change_stream_state import CodeChangeStreamState
from sweepai.dataclasses.file_modification_state import FileModificationState
from sweepai.core.entities import organize_snippets
from sweepai.modify.quick_modify import format_changes
from sweepai.modify.modify_utils import find_best_matches
from sweepai.modify.validate.code_validators import (
    remove_placeholder_comments_from_code,
)
from sweepai.o11y.event_logger import posthog
from sweepai.o11y.log_utils import log_to_file
from sweepai.o11y.posthog_trace import posthog_trace
from sweepai.search.agent.search_prompts import (
    o1_preview_fix_ci_message,
    anthropic_solve_problem_format_message,
    anthropic_conclusion_format,
    anthropic_fix_ci_conclusion_prompt,
    anthropic_fix_ci_message,
    anthropic_fix_ci_selection_prompt,
    anthropic_follow_up_prefix,
    anthropic_format_message,
    anthropic_system_message,
    assistant_prefill_message_for_code_suggestions,
    relevant_snippets_message,
    search_agent_existing_patterns_prompt,
    search_agent_solution_prompt,
)
from sweepai.utils.str_utils import (
    extract_fenced_code_blocks,
    extract_object_fields_from_string,
    extract_xml_tag,
    get_hash,
)
from sweepai.utils.streamable_functions import streamable


def preprocess_messages(messages):
    code_entity_relationship_map = ""
    search_agent_solution = ""
    existing_patterns = ""
    search_solution_message = messages[-1]
    if search_solution_message.role == "function" and search_solution_message.function_call:
        code_entity_relationship_map_dict, failed, _ = extract_object_fields_from_string(
            search_solution_message.function_call.get("thinking", ""),
            ["code_entity_relationship_map"],
        )
        if not failed:
            code_entity_relationship_map = code_entity_relationship_map_dict.get("code_entity_relationship_map", "").strip()  # type: ignore
        else:  # if code_entity_relationship_map is not fetched it may not be in the last message but the ones before, fetch the latest code_entity_relationship_map
            # find the latest code_entity_relationship_map
            for message in reversed(messages):
                if message.role == "function" and message.function_call:
                    code_entity_relationship_map_dict, failed, _ = extract_object_fields_from_string(
                        message.function_call.get("thinking", ""),
                        ["code_entity_relationship_map"],
                    )
                    if not failed:
                        code_entity_relationship_map = code_entity_relationship_map_dict.get("code_entity_relationship_map", "").strip()  # type: ignore
                    if code_entity_relationship_map:
                        break
        try:
            search_agent_solution = (
                search_solution_message.function_call.get("function_parameters", {}).get("solution", "").strip()
            )
        except Exception as e:
            logger.warning(f"Error during search agent solution extraction: {e}")
            search_agent_solution = ""
        try:
            existing_patterns = (
                search_solution_message.function_call.get("function_parameters", {})
                .get("existing_patterns", "")
                .strip()
            )
        except Exception as e:
            logger.warning(f"Error during existing patterns extraction: {e}")
            existing_patterns = ""
    return (
        code_entity_relationship_map,
        search_agent_solution,
        existing_patterns,
        messages,
    )


# TODO: port the tracking ID over to the chat agent
@streamable
@posthog_trace
def chat_agent(
    username: str,
    cloned_repo: ClonedRepo,
    messages: list[Message],
    modify_files_dict: dict,
    snippets: list[Snippet],
    context: ChatAgentContext,
    metadata: dict = {},
):
    tracking_id = metadata.get("tracking_id", get_hash())
    cleaned_repo_name = cloned_repo.repo_full_name.replace("/", "--")
    file_identifier = f"chat-agent-{username}-{cleaned_repo_name}-{tracking_id}"
    logger.info(f"Logs redirected to {file_identifier}")
    with log_to_file(file_identifier):
        model = LATEST_CLAUDE_SONNET_MODEL
        # model = "gpt-4o"
        if not snippets:
            raise ValueError("No snippets were sent.")
        repo_name = cloned_repo.repo_full_name
        system_message = (
            anthropic_system_message.render(use_case=context.use_case)
        )  # TODO: handle this better
        # go backward through the messages until we get a user message with role == 'user'
        initial_user_message = ""
        for message in reversed(messages):
            if message.role == "user":
                initial_user_message = message.content or ""
                break
        if initial_user_message == "":
            raise ValueError("Something went wrong! No user message was found in the messages.")
        thread: Thread = Thread.from_system_message_string(prompt_string=system_message)
        # make a copy of the messages to avoid modifying the original

        pr_snippets = []
        # remove function calls for the following tools
        messages = list(
            filter(
                lambda message: not (
                    (
                        message.role == "function"
                        and message.function_call.get("function_name", "")
                        in (
                            "vector_search",
                            "add_files_to_context",
                            "done_file_search",
                            "access_file",
                            "ripgrep",
                        )
                    )
                    or (hasattr(message, "content") and message.content.strip() == "")
                ),
                messages,
            )
        )
        # parsing logic for funtion calls
        all_pulls_messages = ""
        pr_snippets_for_rendering_ci_prompts = []
        for message in messages:
            if message.role == "user":
                message.content = message.content.strip()
            elif message.role == "function":
                message.role = "user"
                function_name = message.function_call.get("function_name", "")
                if function_name == "done_file_search":
                    message.content = message.function_call.get("function_parameters", {}).get("solution", "")
                elif function_name == "ripgrep":
                    message.content = message.content.split("First, think step-by-step in a scratchpad")[0]
                    message.content = message.content.replace("<ripgrep_response>", "")
            if message.function_call:
                message.function_call = None  # pass certain validations
            if message.annotations:
                new_pr_snippets, skipped_pr_snippets, pulls_messages = get_pr_snippets(
                    repo_name,
                    message.annotations,
                    cloned_repo,
                    context=context,
                )
                pr_snippets_for_rendering_ci_prompts = new_pr_snippets + skipped_pr_snippets
                pr_snippets = organize_snippets(pr_snippets + new_pr_snippets)
                if pulls_messages:
                    all_pulls_messages += pulls_messages.rstrip() + "\n\n"
                    message.content += (
                        "\n\nPull requests:\n"
                        + pulls_messages
                        + f"\n\nBe sure to summarize the contents of the pull request during the analysis phase separately from other relevant files.\n\nRemember, the user's request was:\n\n<message>\n{message.content}\n</message>"
                    )

        relevant_pr_snippets: list[Snippet] = []
        other_relevant_snippets: list[Snippet] = []
        if pr_snippets:
            for snippet in snippets:
                if snippet.file_path in [pr_snippet.file_path for pr_snippet in pr_snippets]:
                    relevant_pr_snippets.append(snippet)
                else:
                    other_relevant_snippets.append(snippet)

        snippets_message = relevant_snippets_message.render(
            repo_name=repo_name,
            relevant_snippets=snippets,
            relevant_pr_snippets=relevant_pr_snippets,
            other_relevant_snippets=other_relevant_snippets,
        )

        thread.messages = [
            SystemMessage(content=thread.messages[0].content),
            UserMessage(content=snippets_message),
        ]

        if context.use_case == "chat":
            thread.messages.extend([message for message in messages[:-1] if message.content != "Loading..."])

        if all_pulls_messages:
            thread.messages.append(
                UserMessage(
                    content=all_pulls_messages,
                )
            )

        search_agent_solution = metadata.get("search_agent_solution", "")
        code_entity_relationship_map = metadata.get("code_entity_relationship_map", "")
        existing_patterns = metadata.get("existing_patterns", "")
        # append the latest search solution message to the end of the messages
        if search_agent_solution or code_entity_relationship_map:
            formatted_search_agent_prompt = ""
            if search_agent_solution:
                formatted_search_agent_prompt += search_agent_solution_prompt.format(
                    search_agent_solution=search_agent_solution
                )
            # unsure how to solve for when the code_entity_relationship_map is unchanged probably need to check when metadata is entered
            if existing_patterns:
                formatted_search_agent_prompt += search_agent_existing_patterns_prompt.format(
                    existing_patterns=existing_patterns
                )
            thread.messages.append(UserMessage(content=formatted_search_agent_prompt, key="search_agent_solution"))

        thread.messages.append(
            UserMessage(
                content=anthropic_follow_up_prefix.render(
                    user_request=initial_user_message,
                    use_case=context.use_case,
                    history=context.history,
                )
            )
        )

        rewritten_solution = ""
        if context.use_case == "ci":
            # get the assistant's reply with the problem solving format (automatically appended to the thread)
            # this is also a great spot to do validation via another LLM call
            if model.startswith("o1-"):
                __o1_preview_fix_ci_message_response = continuous_llm_calls(
                    username,
                    thread,
                    content=o1_preview_fix_ci_message.render(
                        pr_files=pr_snippets_for_rendering_ci_prompts,
                        use_case=context.use_case,
                        history=context.history,
                    ),
                    stop_sequences=["</analysis>"],
                    model="o1-preview",
                )

                anthropic_fix_ci_conclusion_prompt_response = continuous_llm_calls(
                    username,
                    thread,
                    content=anthropic_fix_ci_conclusion_prompt,
                    stop_sequences=["</rewritten_solution>"],
                    model=model,
                )
                rewritten_solution = extract_xml_tag(anthropic_fix_ci_conclusion_prompt_response, "rewritten_solution", include_closing_tag=False)
                rewritten_solution = remove_placeholder_comments_from_code(
                    rewritten_solution
                )
                logger.info(f"REWRITTEN SOLUTION:\n{rewritten_solution}")

                thread.messages = thread.messages[:-4]
                # breakpoint()
            else:
                __anthropic_fix_ci_message_response = continuous_llm_calls(
                    username,
                    thread,
                    content=anthropic_fix_ci_message.render(
                        pr_files=pr_snippets_for_rendering_ci_prompts,
                        use_case=context.use_case,
                        history=context.history,
                    ),
                    stop_sequences=["</analysis>", "</format>"],
                    model=model,
                )
                logger.debug(f"Suggestions: {__anthropic_fix_ci_message_response}")
                __anthropic_fix_ci_selection_prompt_response = continuous_llm_calls(
                    username,
                    thread,
                    content=anthropic_fix_ci_selection_prompt.render(
                        pr_files=pr_snippets_for_rendering_ci_prompts,
                        use_case=context.use_case,
                        history=context.history,
                    ),
                    stop_sequences=["</best_fix_selection>", "</format>"],
                    model=model,
                )
                logger.debug(f"Selection: {__anthropic_fix_ci_selection_prompt_response}")
                anthropic_fix_ci_conclusion_prompt_response = continuous_llm_calls(
                    username,
                    thread,
                    content=anthropic_fix_ci_conclusion_prompt,
                    stop_sequences=["</rewritten_solution>"],
                    model=model,
                )
                rewritten_solution = extract_xml_tag(anthropic_fix_ci_conclusion_prompt_response, "rewritten_solution", include_closing_tag=False)
                rewritten_solution = remove_placeholder_comments_from_code(
                    rewritten_solution
                )  # remove placeholder comments from here only
                # remove the fix_ci_message and fix_ci_selection_prompt from the thread
                logger.info(f"REWRITTEN SOLUTION:\n{rewritten_solution}")
                # breakpoint() # Really good spot to put a breakpoint
                # Delete 6 because we have 3 calls * 2 messages, this resets the context
                thread.messages = thread.messages[:-6]
        elif False: # experimental
            __anthropic_solve_problem_format_message_response = continuous_llm_calls(
                username,
                thread,
                content=anthropic_solve_problem_format_message.render(use_case=context.use_case),
                model="o1-preview",
            )
            rewritten_solution = remove_placeholder_comments_from_code(rewritten_solution)
            logger.info(f"REWRITTEN SOLUTION:\n{rewritten_solution}")

        user_message = (
            anthropic_format_message.render(use_case=context.use_case)
            + "\n\n"
            + anthropic_conclusion_format.render(user_request=initial_user_message, proposed_solution=rewritten_solution)
        )
        new_messages: list[Message] = []
        # convert this to a FileModificationState
        code_stream_modify_files_dict = FileModificationState()
        previous_modify_files_dict: dict = metadata.get("modify_files_dict", {})
        for file_path, file_content in previous_modify_files_dict.items():
            code_stream_modify_files_dict.add_or_update_file(file_path, file_content["contents"], file_content["contents"])

        modify_files_dict: FileModificationState = FileModificationState.from_dict(previous_modify_files_dict)

        # use this to manage the state of the stream when generating code suggestions
        code_change_stream_state: CodeChangeStreamState = CodeChangeStreamState(
            username=username,
            model=model,
            cloned_repo=cloned_repo,
            thread=thread,
            stop_sequences=["</user_response>"],
            modify_files_dict=code_stream_modify_files_dict,
        )

        yield new_messages

        request_assistant_message_content = assistant_prefill_message_for_code_suggestions.render(context=context)
        user_response = ""
        result_string = ""
        current_messages = []
        for result_string in code_change_stream_state.stream(
            user_message,
            request_assistant_message_content,
        ):
            if len(result_string) < 200:  # preamble
                continue
            code_change_stream_state.update_state_with_new_response(result_string)
            analysis = extract_xml_tag(result_string, "analysis", include_closing_tag=False) or ""
            user_response = extract_xml_tag(result_string, "user_response", include_closing_tag=False) or ""
            if analysis or user_response:
                current_messages = []

                if analysis:
                    current_messages.append(
                        Message(
                            content=analysis,
                            role="function",
                            function_call={
                                "function_name": "analysis",
                                "function_parameters": {},
                                "is_complete": bool(user_response),
                            },
                        )
                    )

                if user_response:
                    current_messages.append(AssistantMessage(content=user_response, annotations={}))

                yield [*new_messages, *current_messages]
            else:
                current_messages = [AssistantMessage(content=result_string, annotations={})]
                yield [*new_messages, *current_messages]

        new_messages.extend(current_messages)
        if rewritten_solution:
            new_messages[-1].content = (
                "## Thinking\n\n" + rewritten_solution + "\n\n## Suggestion\n\n" + new_messages[-1].content
            )

        yield new_messages

        message_content: str = new_messages[-1].content

        # EXTRACTING CODE ANNOTATIONS
        code_fence_blocks = extract_fenced_code_blocks(message_content)
        snippet_code_matches: dict[str, dict[str, list[dict[str, str]]]] = {}
        for code_content, code_data in code_fence_blocks.items():
            if code_data.get("language", "") != "diff":
                code_matches: dict[str, list[dict[str, str]]] = {}
                # get fuzzy match based on current snippet context
                for snippet in snippets:
                    try:
                        best_matches = find_best_matches(
                            code_content,
                            snippet.content,
                            threshold=90,
                            tokenized=True,
                            context=2,
                        )
                        for match in best_matches:
                            matched_code, score, start, end = match
                            if snippet.file_path not in code_matches:
                                code_matches[snippet.file_path] = []
                            snippet_match = {
                                "code": matched_code,
                                "start": start,
                                "end": end,
                                "score": score,
                                "language": code_data.get("language", ""),
                            }
                            code_matches[snippet.file_path].append(snippet_match)
                    except Exception as e:
                        logger.error(f"Error finding best matches: {e}")
                        continue
                snippet_code_matches[code_content] = code_matches
        if snippet_code_matches:
            if not hasattr(new_messages[-1], "annotations") or new_messages[-1].annotations is None:
                new_messages[-1].annotations = {}
            # Now you can safely assign snippet_code_matches
            new_messages[-1].annotations["snippetCodeMatches"] = snippet_code_matches
            yield new_messages

        code_change_stream_state.apply_last_code_change()
        modify_files_dict = code_change_stream_state.modify_files_dict
        assert len(modify_files_dict) or context.use_case == "chat"

        modify_files_dict = format_changes(modify_files_dict, cloned_repo)
        # # EXTRACTING CODE SUGGESTIONS
        # code_suggestions = extract_and_clean_code_suggestions(message_content, cloned_repo)
        # logger.info(f"Code Suggestions:\n{render_code_suggestions_for_logging(code_suggestions)}")

        # # VALIDATING CODE SUGGESTIONS
        # file_change_requests, current_content = validate_code_suggestions(
        #     code_suggestions, cloned_repo, modify_files_dict, new_messages[-1].content
        # )
        # new_messages[-1].annotations["codeSuggestions"] = code_suggestions
        # new_messages[-1].content = current_content

        # ## NEW MODIFY START
        # parsed_code_suggestions = [CodeSuggestion.from_camel_case(code_suggestion) for code_suggestion in code_suggestions]
        # new_modify_files_dict = quick_modify(parsed_code_suggestions, cloned_repo, modify_files_dict)

        # new_messages[-1].annotations["codeSuggestions"] = new_modify_files_dict.to_raw_code_suggestions()
        new_messages[-1].annotations["codeSuggestions"] = modify_files_dict.to_raw_code_suggestions()

        assert new_messages[-1].annotations["codeSuggestions"] or context.use_case == "chat"
        yield new_messages

        ## MODIFY END

        posthog.capture(
            metadata["username"],
            "chat_codebase complete",
            properties={
                **metadata,
                "messages": [message.model_dump() for message in messages],
            },
        )



from __future__ import annotations

import requests

from pydantic import BaseModel

from sweepai.config.server import GITHUB_GRAPHQL_BASE_URL
from sweepai.core.github_utils import github_integration, get_token

def generate_error_message(errors: list[dict]) -> str:
    error_message = ""
    for error in errors:
        error_message += error.pop("message") + f", details: {error}\n"
    return error_message

class GitHubGraphQLError(Exception):
    message: str

class GitHubGraphQLClient(BaseModel):
    token: str

    @classmethod
    def from_username(cls, username: str) -> GitHubGraphQLClient:
        installation = github_integration.get_user_installation(username=username)
        token = get_token(installation.id)
        return cls(token=token)

    def query(self, query: str, variables: dict) -> dict:
        response = requests.post(
            GITHUB_GRAPHQL_BASE_URL,
            json={"query": query, "variables": variables},
            headers={"Authorization": f"Bearer {self.token}"},
        )
        response.raise_for_status()
        data = response.json()
        if errors := data.get("errors"):
            raise GitHubGraphQLError(generate_error_message(errors))
        return data



from sweepai.utils.format_utils import Prompt

anthropic_follow_up_prefix = Prompt(
    """{{ "The following GitHub Action is failing:" if use_case == "ci" else "Here is the user's request:" }}
{% if history | default('') %}
Here are your previous attempts at fixing the issue:
<fix_attempts>
{{ history }}
</fix_attempts>
{% endif %}
{{ user_request | trim('\n') | wrap_xml("failing_github_actions_logs" if use_case == "ci" else "user_request") }}
"""
)

anthropic_format_message = Prompt(
    """# Guidelines

## Code Changes
{%- if use_case == "chat" %}
1. If the user asks for code changes, you must add <code_change> blocks to the <user_response> block following the format provided below. Reference the proposed solution to the problem.
2. When providing code changes, ALWAYS provide the complete code. Any ellipsis or summary comments like "// Add any mocks here" or "// ... (existing tests)" are not permitted. Look at relevant examples provided. Placeholder comments or empty stub functions are never permitted. Do not provide incomplete code.
{% else %}
1. You must respond with code changes. To do so, you must add <code_change> blocks to the <user_response> block following the format provided below. Reference the proposed solution to the problem.
2. ALWAYS provide complete code. Any ellipsis or summary comments like "// Add any mocks here" or "// ... (existing tests)" are not permitted. Look at relevant examples provided. Placeholder comments or empty stub functions are never permitted.
{% endif -%}
3. To edit multiple parts of an existing file, write separate <code_change> blocks for each part. This is crucial for maintaining clarity and organization in your response.
4. When creating a new code file, you may use a single <code_change> block to create the entire file. Leave the original code empty.

You MUST follow the following XML-based format:

# Format

<user_response>
{%- if use_case == "chat" %}
1. Analyze each of the files provided to understand the codebase and describe how they are related.
2. Determine if the user asked for any changes to the codebase. If not, write a complete response addressing all of the user's queries. Use GitHub Flavored Markdown to format your response for maximal readability. This means using lists, headers, codeblocks and blockquotes extensively.
3. Otherwise, write complete code changes that solve the user's problem in full detail.
{%- else %}
Write complete code changes that solve the user's problem in full detail, addressing all of the user's requests.
{%- endif %}

To suggest code changes:
1. For creating a new file:
    1. Say "Let's create a new file named <file_path> to do xyz"
    2. Write one <code_change> block and write all of the new contents in one block. Drop the <original_code> block and only provide the <new_code> block.
2. If you are adding a new function (like a test case):
    1. Say "Let's add a new function to <file_path> to do xyz"
    2. Find a suitable location in the file to add the new code, such as after an existing method or test case.
    3. Write a <code_change> block and write all of the old code in the <original_code> block. Then, copy the old code, appended by the new code, into the <new_code> block.
3. If you are editing code in a file, list each section of each file that you would like to change, followed by how you will change it:
    1. Describe the section of file and how it will be changed. This will correspond to one <code_change> block.
    [Repeat for each section. To change multiple parts of the file, you MUST write multiple separate <code_change> blocks.]
4. If imports need to be updated in any of the code changes and they have not been added yet, create a separate <code_change> block for the imports, and make this change at the very end.
{% if use_case == "chat" %}
When providing code changes, remember to always provide complete code. Placeholder comments, ellipsis, or empty stub functions such as "// Add any mocks here" or "// ... (existing tests)" are not permitted.
{% else %}
Remember to always provide complete code. Placeholder comments, ellipsis, or empty stub functions such as "// Add any mocks here" or "// ... (existing tests)" are not permitted.
{% endif %}
Use the following format for each <code_change> block:

<code_change>
<file_path>
path/to/file.py
</file_path>
<justification>
Explain what you will change, e.g. "I will add a new method to do xyz." or "I will edit foo to support xyz."
Repeat the following phrase after your explanation: "I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task."
</justification>
<original_code>
```
(Optional) Original code from the file that new_code will replace. Leave empty for writing new files or adding new functions.
```
</original_code>
<new_code>
```
Place the new code here.
```
</new_code>
</code_change>

If you are editing a block of code, provide the <original_code> block.
If you are adding a new function or test case, do not provide the <original_code> block.
</user_response>

### Examples

Here are a few illustrative examples of how to use the <code_change> format in different scenarios:

<examples>
### Example 1: Modifying a line of code

<example_user_request>
In the Rectangle class, the area calculation is incorrect. Can you fix it?
</example_user_request>

<code_change>
<file_path>
src/utils/Rectangle.java
</file_path>
<justification>
I will correct the calculation of the area to use multiplication instead of addition.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
public double area() {
    return width + height;
}
```
</original_code>
<new_code>
```
public double area() {
    return width * height;
}
```
</new_code>
</code_change>

### Example 2: Adding new code to an existing file, such as a new method or a new test case

<example_user_request>
Can you add a method to calculate the median in our DataProcessor class?
</example_user_request>

<code_change>
<file_path>
src/main/java/com/example/DataProcessor.java
</file_path>
<justification>
I will add a new method 'calculateMedian' to the DataProcessor class to calculate the median of a list of numbers. I will place this method after the processData method.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
    public Map<String, Object> processData(List<Double> dataList) {
        if (dataList.isEmpty()) {
            return Collections.singletonMap("error", "Empty input list");
        }

        Map<String, Object> result = new HashMap<>();

        // Calculate basic statistics
        result.put("count", dataList.size());
        double sum = dataList.stream().mapToDouble(Double::doubleValue).sum();
        result.put("sum", sum);
        double mean = sum / dataList.size();
        result.put("mean", mean);

        // Calculate min and max
        result.put("min", Collections.min(dataList));
        result.put("max", Collections.max(dataList));

        // Calculate variance and standard deviation
        double variance = dataList.stream()
                .mapToDouble(x -> Math.pow(x - mean, 2))
                .average()
                .orElse(Double.NaN);
        result.put("variance", variance);
        result.put("std_dev", Math.sqrt(variance));

        // Normalize the data
        double min = (double) result.get("min");
        double max = (double) result.get("max");
        List<Double> normalized = dataList.stream()
                .map(x -> (x - min) / (max - min))
                .collect(Collectors.toList());
        result.put("normalized", normalized);

        // Round all double values to 2 decimal places
        result.replaceAll((k, v) -> {
            if (v instanceof Double) {
                return Math.round((Double) v * 100.0) / 100.0;
            } else if (v instanceof List) {
                return ((List<?>) v).stream()
                        .map(x -> x instanceof Double ? Math.round((Double) x * 100.0) / 100.0 : x)
                        .collect(Collectors.toList());
            }
            return v;
        });

        return result;
    }
```
</original_code>
<new_code>
```
    public Map<String, Object> processData(List<Double> dataList) {
        if (dataList.isEmpty()) {
            return Collections.singletonMap("error", "Empty input list");
        }

        Map<String, Object> result = new HashMap<>();

        // Calculate basic statistics
        result.put("count", dataList.size());
        double sum = dataList.stream().mapToDouble(Double::doubleValue).sum();
        result.put("sum", sum);
        double mean = sum / dataList.size();
        result.put("mean", mean);

        // Calculate min and max
        result.put("min", Collections.min(dataList));
        result.put("max", Collections.max(dataList));

        // Calculate variance and standard deviation
        double variance = dataList.stream()
                .mapToDouble(x -> Math.pow(x - mean, 2))
                .average()
                .orElse(Double.NaN);
        result.put("variance", variance);
        result.put("std_dev", Math.sqrt(variance));

        // Normalize the data
        double min = (double) result.get("min");
        double max = (double) result.get("max");
        List<Double> normalized = dataList.stream()
                .map(x -> (x - min) / (max - min))
                .collect(Collectors.toList());
        result.put("normalized", normalized);

        // Round all double values to 2 decimal places
        result.replaceAll((k, v) -> {
            if (v instanceof Double) {
                return Math.round((Double) v * 100.0) / 100.0;
            } else if (v instanceof List) {
                return ((List<?>) v).stream()
                        .map(x -> x instanceof Double ? Math.round((Double) x * 100.0) / 100.0 : x)
                        .collect(Collectors.toList());
            }
            return v;
        });

        return result;
    }
    public double calculateMedian(List<Double> dataList) {
        System.out.println("Calculating median");
        if (dataList.isEmpty()) {
            throw new IllegalArgumentException("Cannot calculate median of an empty list");
        }
        
        PriorityQueue<Double> minHeap = new PriorityQueue<>();
        PriorityQueue<Double> maxHeap = new PriorityQueue<>(Collections.reverseOrder());
        
        for (double num : dataList) {
            if (maxHeap.isEmpty() || num < maxHeap.peek()) {
                maxHeap.offer(num);
            } else {
                minHeap.offer(num);
            }
            
            if (maxHeap.size() > minHeap.size() + 1) {
                minHeap.offer(maxHeap.poll());
            } else if (minHeap.size() > maxHeap.size()) {
                maxHeap.offer(minHeap.poll());
            }
        }
        
        if (dataList.size() % 2 == 0) {
            return (maxHeap.peek() + minHeap.peek()) / 2.0;
        } else {
            return maxHeap.peek();
        }
    }
```
</new_code>
</code_change>

<code_change>
<file_path>
src/main/java/com/example/DataProcessor.java
</file_path>
<justification>
I will add the imports for PriorityQueue for the new method 'calculateMedian'.
</justification>
<original_code>
```
import java.util.List;
import java.util.ArrayList;
```
</original_code>
<new_code>
```
import java.util.List;
import java.util.ArrayList;
import java.util.PriorityQueue;
```
</new_code>
</code_change>

### Example 3: Creating a new file

<example_user_request>
Can you create a new unit test file for our MathOperations class?
</example_user_request>

<code_change>
<file_path>
src/test/java/com/example/MathOperationsTest.java
</file_path>
<justification>
I will create a new unit test file for the MathOperations class.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<new_code>
```
package com.example;

import org.junit.jupiter.api.Test;
import static org.junit.jupiter.api.Assertions.*;

class MathOperationsTest {

    @Test
    void testAdd() {
        assertEquals(7, MathOperations.add(3, 4), "3 + 4 should equal 7");
        assertEquals(0, MathOperations.add(-1, 1), "-1 + 1 should equal 0");
    }

    @Test
    void testSubtract() {
        assertEquals(5, MathOperations.subtract(10, 5), "10 - 5 should equal 5");
        assertEquals(0, MathOperations.subtract(-1, -1), "-1 - (-1) should equal 0");
    }

    @Test
    void testMultiply() {
        assertEquals(12, MathOperations.multiply(3, 4), "3 * 4 should equal 12");
        assertEquals(-15, MathOperations.multiply(3, -5), "3 * -5 should equal -15");
        assertEquals(0, MathOperations.multiply(100, 0), "100 * 0 should equal 0");
    }

    @Test
    void testDivide() {
        assertEquals(2.5, MathOperations.divide(5, 2), 0.0001, "5 / 2 should equal 2.5");
        assertEquals(-2, MathOperations.divide(10, -5), "10 / -5 should equal -2");
        assertThrows(ArithmeticException.class, () -> MathOperations.divide(10, 0), "Division by zero should throw ArithmeticException");
    }
}
```
</new_code>
</code_change>

### Example 4: Handling changes across multiple functions

<example_user_request>
In our UserService class, we need to add logging for all public methods. Can you add a log statement at the beginning of each public method that logs the method name and parameters? Use the following format:
logger.info("Entering method: methodName with parameters: param1={}, param2={}", param1, param2);
// ... do this for all public methods
</example_user_request>

First, let's add the logger and modify the class declaration:

<code_change>
<file_path>
src/main/java/com/example/service/UserService.java
</file_path>
<justification>
I will add a logger to the UserService class for logging method entries.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
@Service
public class UserService {

    private final UserRepository userRepository;

    @Autowired
    public UserService(UserRepository userRepository) {
        this.userRepository = userRepository;
    }
```
</original_code>
<new_code>
```
@Service
public class UserService {

    private static final Logger logger = LoggerFactory.getLogger(UserService.class);

    private final UserRepository userRepository;

    @Autowired
    public UserService(UserRepository userRepository) {
        this.userRepository = userRepository;
    }
```
</new_code>
</code_change>

Now, let's add logging to the getAllUsers method:

<code_change>
<file_path>
src/main/java/com/example/service/UserService.java
</file_path>
<justification>
Adding logging to the getAllUsers method.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
    public List<User> getAllUsers() {
        return userRepository.findAll();
    }
```
</original_code>
<new_code>
```
    public List<User> getAllUsers() {
        logger.info("Entering method: getAllUsers");
        return userRepository.findAll();
    }
```
</new_code>
</code_change>

Let's add logging to the getUserById method:

<code_change>
<file_path>
src/main/java/com/example/service/UserService.java
</file_path>
<justification>
Adding logging to the getUserById method.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
    public Optional<User> getUserById(Long id) {
        return userRepository.findById(id);
    }
```
</original_code>
<new_code>
```
    public Optional<User> getUserById(Long id) {
        logger.info("Entering method: getUserById with parameters: id={}", id);
        return userRepository.findById(id);
    }
```
</new_code>
</code_change>

Now, let's add logging to the createUser method:

<code_change>
<file_path>
src/main/java/com/example/service/UserService.java
</file_path>
<justification>
Adding logging to the createUser method.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
    public User createUser(User user) {
        return userRepository.save(user);
    }
```
</original_code>
<new_code>
```
    public User createUser(User user) {
        logger.info("Entering method: createUser with parameters: user={}", user);
        return userRepository.save(user);
    }
```
</new_code>
</code_change>

Let's add logging to the updateUser method:

<code_change>
<file_path>
src/main/java/com/example/service/UserService.java
</file_path>
<justification>
Adding logging to the updateUser method.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
    public User updateUser(Long id, User userDetails) {
        User user = userRepository.findById(id)
                .orElseThrow(() -> new RuntimeException("User not found with id: " + id));

        user.setName(userDetails.getName());
        user.setEmail(userDetails.getEmail());

        return userRepository.save(user);
    }
```
</original_code>
<new_code>
```
    public User updateUser(Long id, User userDetails) {
        logger.info("Entering method: updateUser with parameters: id={}, userDetails={}", id, userDetails);
        User user = userRepository.findById(id)
                .orElseThrow(() -> new RuntimeException("User not found with id: " + id));

        user.setName(userDetails.getName());
        user.setEmail(userDetails.getEmail());

        return userRepository.save(user);
    }
```
</new_code>
</code_change>

Finally, let's add logging to the deleteUser method:

<code_change>
<file_path>
src/main/java/com/example/service/UserService.java
</file_path>
<justification>
Adding logging to the deleteUser method.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
    public void deleteUser(Long id) {
        userRepository.deleteById(id);
    }
```
</original_code>
<new_code>
```
    public void deleteUser(Long id) {
        logger.info("Entering method: deleteUser with parameters: id={}", id);
        userRepository.deleteById(id);
    }
```
</new_code>
</code_change>

Lastly, let's update the imports:

<code_change>
<file_path>
src/main/java/com/example/service/UserService.java
</file_path>
<justification>
Adding the necessary import for the Logger.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
import com.example.model.User;
import com.example.repository.UserRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.Optional;
```
</original_code>
<new_code>
```
import com.example.model.User;
import com.example.repository.UserRepository;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.Optional;
```
</new_code>
</code_change>
{%- if use_case != "chat" %}

This example demonstrates how to break down a single change (adding logging to all public methods) into multiple focused <code_change> blocks.

### Example 5: Creating a completely new file

<example_user_request>
We need to create a new utility class called StringUtils that provides common string manipulation methods. Can you create this class with methods for reversing a string, checking if a string is a palindrome, and counting the occurrences of a substring?
</example_user_request>

<code_change>
<file_path>
src/main/java/com/example/utils/StringUtils.java
</file_path>
<justification>
I will create a new utility class StringUtils with common string manipulation methods to improve code reusability and maintainability.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<new_code>
```
package com.example.utils;

public class StringUtils {

    /**
     * Reverses the given string.
     *
     * @param str the string to reverse
     * @return the reversed string
     */
    public static String reverse(String str) {
        if (str == null || str.isEmpty()) {
            return str;
        }
        return new StringBuilder(str).reverse().toString();
    }

    /**
     * Checks if the given string is a palindrome.
     * A palindrome reads the same backwards as forwards, ignoring case and non-alphanumeric characters.
     *
     * @param str the string to check
     * @return true if the string is a palindrome, false otherwise
     */
    public static boolean isPalindrome(String str) {
        if (str == null || str.isEmpty()) {
            return true;
        }
        String cleaned = str.replaceAll("[^a-zA-Z0-9]", "").toLowerCase();
        return cleaned.equals(new StringBuilder(cleaned).reverse().toString());
    }

    /**
     * Counts the occurrences of a substring in a given string.
     *
     * @param str the string to search in
     * @param subStr the substring to search for
     * @return the number of occurrences of subStr in str
     */
    public static int countOccurrences(String str, String subStr) {
        if (str == null || str.isEmpty() || subStr == null || subStr.isEmpty()) {
            return 0;
        }
        int count = 0;
        int lastIndex = 0;
        while (lastIndex != -1) {
            lastIndex = str.indexOf(subStr, lastIndex);
            if (lastIndex != -1) {
                count++;
                lastIndex += subStr.length();
            }
        }
        return count;
    }
}
```
</new_code>
</code_change>

### Example 6: Deleting sections of a file

<example_user_request>
We need to remove the deprecated methods from our LegacyUtils class. Can you delete the outdated methods and their associated comments?
</example_user_request>

<code_change>
<file_path>
src/main/java/com/example/utils/LegacyUtils.java
</file_path>
<justification>
Removing the deprecated max and toLowerCase methods to clean up the codebase and reduce maintenance overhead.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
    /**
     * @deprecated Use java.lang.Math.max() instead.
     * Returns the maximum of two integers.
     * @param a First integer
     * @param b Second integer
     * @return The maximum of a and b
     */
    @Deprecated
    public static int max(int a, int b) {
        return a > b ? a : b;
    }

    /**
     * @deprecated Use java.lang.String.toLowerCase() instead.
     * Converts a string to lowercase.
     * @param str Input string
     * @return Lowercase version of the input string
     */
    @Deprecated
    public static String toLowerCase(String str) {
        return str.toLowerCase();
    }
}
```
</original_code>
<new_code>
```
```
</new_code>
</code_change>

When deleting sections of a file, you can leave <new_code> empty. Do not leave behind any placeholder comments.
                                  
### Example 7: Making multiple changes to a file

<example_user_request>
We need to refactor our UserController class to implement pagination and add input validation.

1. Modify the `UserController` class to implement pagination and add input validation.
2. Add input validation to the `createUser` method.
3. Update the imports to include the new classes and annotations used in the refactored code.
</example_user_request>

Let's modify the `UserController` class to implement pagination and add input validation. We'll make these disjoint changes in multiple code blocks for clarity.
First, let's implement pagination for the `getAllUsers` method.

<code_change>
<file_path>
src/main/java/com/example/controller/UserController.java
</file_path>
<justification>
Implementing pagination for the getAllUsers method to improve performance and reduce memory usage when dealing with large datasets.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
@GetMapping
public ResponseEntity<List<User>> getAllUsers() {
    List<User> users = userService.getAllUsers();
    return ResponseEntity.ok(users);
}
```
</original_code>
<new_code>
```
@GetMapping
public ResponseEntity<Page<User>> getAllUsers(
        @RequestParam(defaultValue = "0") int page,
        @RequestParam(defaultValue = "10") int size) {
    Pageable pageable = PageRequest.of(page, size);
    Page<User> users = userService.getAllUsers(pageable);
    return ResponseEntity.ok(users);
}
```
</new_code>
</code_change>

Now let's add input validation to the `createUser` method.
<code_change>
<file_path>
src/main/java/com/example/controller/UserController.java
</file_path>
<justification>
Adding input validation to the createUser method.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
@PostMapping
public ResponseEntity<User> createUser(@RequestBody User user) {
    User createdUser = userService.createUser(user);
    return ResponseEntity.status(HttpStatus.CREATED).body(createdUser);
}
```
</original_code>
<new_code>
```
@PostMapping
public ResponseEntity<User> createUser(@Valid @RequestBody User user) {
    User createdUser = userService.createUser(user);
    return ResponseEntity.status(HttpStatus.CREATED).body(createdUser);
}
```
</new_code>
</code_change>

Now let's update the imports to include the new classes and annotations used in the refactored code.
<code_change>
<file_path>
src/main/java/com/example/controller/UserController.java
</file_path>
<justification>
Updating the imports to include the new classes and annotations used in the refactored code.
I will not be lazy when generating the original_code and new_code blocks and include all lines of code necessary to solve the task.
</justification>
<original_code>
```
import com.example.model.User;
import com.example.service.UserService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

import java.util.List;
```
</original_code>
<new_code>
```
import com.example.model.User;
import com.example.service.UserService;
import com.example.exception.ResourceNotFoundException;
import com.example.dto.ErrorResponse;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.validation.FieldError;

import javax.validation.Valid;
import java.util.HashMap;
import java.util.Map;
```
</new_code>
</code_change>
{%- endif %}
</examples>

The above examples are for illustrative purposes only. Tailor your response to the user's request.{{ " You will only need to make code changes if the user asks for them." if use_case == "chat" else "" }}"""
)

anthropic_conclusion_format = Prompt(
    """Here is the user's request{{ " (for context)" if proposed_solution.strip() else "" }}:

<user_request>
{{ user_request.strip("\n") }}
</user_request>
{% if proposed_solution.strip() %}
Implement the proposed solution to the problem:

<proposed_solution>
{{proposed_solution}}
</proposed_solution>
{% endif %}"""
)

anthropic_solve_problem_format_message = Prompt(
    """\
# Response Format

Respond in the following format:

## File & User Request Examination

Read all of the above information to build a solution to the problem.
Identify how all of the pieces connect together.
{% if use_case == "chat" %}
## Resolve the User's Request:
Completely resolve the user's request, based on the examination earlier.

## Requirements:
- Focus on the specific files and functions mentioned in the user request. Do not modify other parts of the codebase unless necessary for the requested changes to work.
- Maintain all existing comments in the code unless they directly conflict with the requested changes. If a comment needs to be modified, update it to reflect the new code accurately.
- You may create multiple code changes for each file. 
- Each code change should be in its own block, unless they are contiguous changes in the same file.
- For each significant change, provide a brief explanation of why the change was made and how it addresses the user's request.
{% else %}
## Implement the User's Request:
Completely implement the user's request.

## Requirements:
- Determine the most effective way to implement the user's request and best integrate it with the existing codebase.
- Focus on the specific files and functions mentioned in the user request. Do not modify other parts of the codebase unless necessary for the requested changes to work.
- Maintain all existing comments in the code unless they directly conflict with the requested changes. If a comment needs to be modified, update it to reflect the new code accurately.
- You may create multiple code changes for each file. 
- Each code change should be in its own block, unless they are contiguous changes in the same file.
- For each significant change, provide a brief explanation of why the change was made and how it addresses the user's request.
{% endif %}
## Format:
Use the following formatting for each code change.

List all changes you would like to make to the codebase at a high level.

1. Show the new code.

[file_path]:
```
[the new code]
```

2. For each changed file, show all of the imports needed for the new code.
[file_path]
```
[any imports needed for the new code]
```

Keep making additional code changes until you have satisfied the user's request."""
)

o1_preview_fix_ci_message = Prompt(
    """# Task
Make changes to the code in your pull request to fix the failing GitHub Actions. The codebase was error-free prior to these changes, so these errors are from your pull request.

Fix the failing CI by identifying the root cause of the failure and making the necessary code changes. We must adhere to the original intention of the pull request.

These were the files that were changed in your pull request (the source of the errors):
<pull_request_changed_files>
{{ pr_files | map(attribute='file_path') | list }}
</pull_request_changed_files>

Identify a fix to *your code* (in the files listed above) in the pull request to fix the failing CI. Do not edit files unchanged by your pull request. These must still adhere to the original intention of the pull request. Explain your reasoning for your changes.
"""
)

# This one is starting to become overloaded
anthropic_fix_ci_message = Prompt(
    """# Task
Make changes to the code in your pull request to fix the failing GitHub Actions. The codebase was error-free prior to these changes, so these errors are from your pull request.

Fix the failing CI by identifying the root cause of the failure and making the necessary code changes. We must adhere to the original intention of the pull request.

These were the files that were changed in your pull request (the source of the errors):
<pull_request_changed_files>
{{ pr_files | map(attribute='file_path') | list }}
</pull_request_changed_files>

<format>
<summary>
1. Indicate which files were not changed by your pull request. Say "These files are not the root cause of the error".
2. Indicate what you changed in this pull request. State explicitly "I have made incorrect assumptions" such as how to use certain interfaces in this codebase, class methods and attributes, or what features are available. For example, your unit test's intent or interface usage is incorrect.
3. Make a bulleted list of all previous attempts at addressing this error and any learnings from them.
</summary>

<analysis>
1. Read each and every provided pr_file and other_relevant_file to highlight the lines which cause or are related to the error.

<relevant_file index="0">
[ file path ]
</relevant_file>
<highlighted_relevant_lines>
```
[ Copy the lines which cause or are related to the error. Include < 50 lines per file. ]
```
</highlighted_relevant_lines>

[ Repeat this for each file. ]

2. Then, for each relevant variable, method or function, indicate their respective types or interfaces in this format:

<interfaces>
foo: dict[str, int]
bar(x: int) -> int
[ repeat for each relevant variable, method or function ]
</interfaces>

3. Analyze the output errors. Identify the traceback, as well as any relevant debug logs. Also indicate what you have previously tried.

4. Identify potential fixes to the failing CI. These must still adhere the original intention of the pull request. The CI failure is due to a logical error in your changes. List at least 5 potential fixes in your code.

<potential_fix fix_index="i">
[ Clearly describe the root cause and file(s) in which it occurs. The fix must mention specific file(s). Reference the source code to back up your analysis. ]
<fix>
[ Describe the root cause and corresponding fix. Be specific about why it fixes the root cause. ]
<code file_path=[ file path ]>
[ Write sample code to show the root cause and corresponding fix. ]
</code>
</fix>
</potential_fix>
</analysis>
</format>"""
)

# maybe just allow the debug statements
# We previously had <conclusion> but we moved it because the LLM changes weren't granular enough
anthropic_fix_ci_selection_prompt = Prompt(
    """Select the most likely potential mistake. Recall your pull request edited the following files:
{{ pr_files | map(attribute='file_path') | list }}

Guidelines:
- Prefer fixes that involve making changes to the files above.
- Your changes have introduced errors in the codebase.
- Your pull request may have made incorrect assumptions about how the existing code works, how it should be used, or features that are available. Certain attributes or methods may have been used in unintended ways.
- Analyze the source code for answers instead of making assumptions.
- If you're uncertain about what certain variables are set to, add debug statements to the tests.

Use the following format:

<format>
<summary>
1. Indicate which files were not changed by your pull request. Say "These files are not the root cause of the error".
2. Indicate what you changed in this pull request. State explicitly "I have made incorrect assumptions" such as how to use certain interfaces in this codebase, class methods and attributes, or what features are available. For example, your unit test's intent or interface usage is incorrect.
{{ "3. Make a bulleted list of all previous attempts at addressing this error and any learnings from them.\n" if history else "" -}}
</summary>

<critique>
For each potential fix look carefully at the source code provided.
[Copy the potential fix here.]
- Pros:
    - Argument for why this is correct. Back up uncertainties by taking a look at the source code and citing relevant lines.
- Cons:
    - Argument for why this may be incorrect or unsuitable. Back up uncertainties by taking a look at the source code and citing relevant lines. If a potential fix would not accomplish the user's goals, then say so. This is not an acceptable fix.
</critique>

<best_fix_selection>
Based on the critique:
1. Identify the most likely mistake made in the pull request. This should be a specific error or misunderstanding evident from the code changes.
2. Copy the best fix or fixes that address this mistake. Read the pr_files and other_relevant_files to select the best fix(es). The best fix must both solve the pull request's original intent and address the root cause of the error.
3. Explain why this fix is the most suitable solution.
</best_fix_selection>
</format>
"""
)

anthropic_fix_ci_conclusion_prompt = """# Rewrite the solution
Now rewrite the solution as completely new instructions to be passed down to a software engineer:

Guidelines:

1. Keep in mind that the software engineer is not aware of the contents of potential_fix, critique, or best_fix_selection.
2. Use numbered lists with entries such as "In FILE, make the following changes." Reference the index of the file from the provided list of files.
3. If there are multiple changes that need to be made in the same file, you must describe them in separate entries.
    a. For example, if you need to change a function and a variable in the same file, describe the changes in separate entries:
        1. In foo_bar.py, change the function foo() to return an integer.
        2. In foo_bar.py, change the variable bar to be a string.
4. Be specific and comprehensive when describing changes. Do not include changes unless you are certain they can/need to be made. Read the corresponding files above before making any changes.

# Format:
<analysis>
[ Use this to determine which changes need to be made. Ensure you follow the guidelines. ]
</analysis>

Rewrite the solution here.
<rewritten_solution>
[ Rewrite the solution here.]
</rewritten_solution>"""

openai_format_message = """You MUST follow the following XML-based format, including <user_response> and </user_respose> tags:

### Format

You must respond with the following two distinct sections:

# 1. Summary and analysis

<analysis>
1. List and summarize each file from the codebase provided that is relevant to the user's question. You may not need to summarize all provided files.

2. List all the requests made by the user.

3. Organize your response to the user into sections. Plan out reasonable headers so that your response is more digestable.

4. Determine if you need to make code changes or not by evaluating the user's latest request. Are they asking you to write code? If yes, then do you need to create a new file or do you need to modify existing files? Provide a high level plan of what code_changes are needed and how many code_changes will be needed. Use only natural language for this part.
</analysis>

# 2. User Response

<user_response>
Write a complete helpful response to the user's question in full detail, addressing all of the user's requests. Make sure this answer is complete and helpful. Provide code examples, explanations and excerpts wherever possible to provide concrete explanations. When showing code examples, only show MINIMAL excerpts of code that address the user's question.

When showing relevant examples of code, only show MINIMAL excerpts of code that address the user's question. Do NOT copy the whole file, but only the lines that are relevant to the user's question.

When suggesting code changes, you add <code_change> blocks inside the <user_response></user_response> tags.
</user_response>"""

# improve these prompts
anthropic_system_message = Prompt(
    """{{"You are an extremely intelligent and thorough coding assistant that writes high quality code to resolve GitHub issues." if not use_case == "ci" else "You are an extremely intelligent coding assistant that identifies and fixes mistakes in your pull request to fix failing CI." }}"""
)

openai_system_message = (
    """You are a helpful assistant that will answer a user's questions about a codebase to resolve their issue. You are provided with a list of relevant code snippets from the codebase that you can refer to. You can use this information to help the user solve their issue. You may also make function calls to retrieve additional information from the codebase. 

# Guidelines

- Focus on providing high-quality explanations. Start with a high-level overview and then dive into the details.
- Only show code as supplementary evidence or to enhance the explanations.
- Always provide the complete code, including all imports and without using any ellipsis or summary comments.
- Use markdown for your responses, using headers where applicable to improve clarity and lists to enumerate examples.
- Wherever possible, you should suggest code changes. To do so, you must add <code_change> blocks to the <user_response> block following the format provided below.
- Code changes must be atomic. Each code change must be in its own block, unless they are contiguous changes in the same file.
- To change multiple parts of the file, write separate <code_change> blocks.
- When writing code changes, look to see if there are any provided pre-existing examples with relevant patterns or structures. If there are you should follow the same patterns or structures in your code changes unless otherwise specified.
- When creating a new code file, use only one code_change to create the entire file.

# <code_change> Format
First, indicate whether you want to modify an existing file or create a new file, then write in the following format:

<code_change>
<file_path>
path/to/file.py
</file_path>
<original_code>
Copy the original section of code from path/to/file.py. This is the section of code that you will change. Paraphrasing, abbreviating the source code, or placeholder comments such as "// rest of code" are NEVER PERMITTED. Leave empty for creating new files.
</original_code>
<new_code>
New code to replace <original_code> with.
</new_code>
</code_change>

Instead of writing "// rest of code" to represent skipping a section of code, make multiple <code_change> blocks for each section of code that you want to change.

Here are a few examples of how to use the <code_change> format in different scenarios:

# Modifying a line of code
In this example the calculation of the area was corrected.

<code_change>
<file_path>
src/utils/rectangle.py
</file_path>
<original_code>
class Rectangle:
    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

    def area(self):
        return self.width + self.height
</original_code>
<new_code>
class Rectangle:
    def __init__(self, width: int, height: int):
        self.width = width
        self.height = height

    def area(self):
        return self.width * self.height
</new_code>
</code_change>

# Creating a completely new file
In this example a new unit test file was created from scratch in a single code change.
Notice how the original_code is empty to signify that this is a new file.

<code_change>
<file_path>
path/to/tests/new_test_file.py
</file_path>
<original_code>
</original_code>
<new_code>
import unittest
from math_operations import add, subtract

class TestMathOperations(unittest.TestCase):
    def test_add(self):
        self.assertEqual(add(3, 4), 7)
        self.assertEqual(add(-1, 1), 0)

    def test_subtract(self):
        self.assertEqual(subtract(10, 5), 5)
        self.assertEqual(subtract(-1, -1), 0)

if __name__ == '__main__':
    unittest.main()
</new_code>
</code_change>

# Appending onto the end of a file
In this example a new function was added to the end of the file.
Notice how the function definition for add_nums appears in both the original_code and new_code sections, this is to signify that the new function is being appended to the end of the file.

<code_change>
<file_path>
src/utils/utils.py
</file_path>
<original_code>
def add_nums(a, b):
    return a + b
</original_code>
<new_code>
def add_nums(a, b):
    return a + b

def multiply_nums(a, b):
    return a * b
</new_code>
</code_change>

"""
    + openai_format_message
)

pr_format = Prompt(
    """
{% if use_case == "ci" %}
You previously wrote this pull request:
{% else %}
Here are the contents of the referenced pull request {{ url }}:
{% endif %}

<pull_request>
{{ title | wrap_xml('title') }}
{{ body | wrap_xml('body') }}
{{ patch | wrap_xml('patch') }}
</pull_request>
"""
)

relevant_snippets_message = Prompt(
    """# Codebase
Repo: {{ repo_name }}
{% if relevant_pr_snippets %}
# Full files from the pull request:
Here are the files from pull request:

<pr_files>
{{ relevant_pr_snippets|render_snippets }}
</pr_files>

Here are other relevant files from the initial search results from the codebase:

<other_relevant_files>
{{ other_relevant_snippets|render_snippets }}
</other_relevant_files>
{% else %}
# Relevant codebase files:
Here are the initial search results from the codebase. These will be your primary reference to solve the problem:

<relevant_files>
{{ relevant_snippets|render_snippets }}
</relevant_files>
{% endif %}
"""
)

search_agent_existing_patterns_prompt = """
# Existing Patterns
Below are a series of existing patterns that the code base uses. Incorporate these patterns as best you can into your solution

<existing_patterns>
{existing_patterns}
</existing_patterns>
"""

search_agent_solution_prompt = """
# Potential Resolution
Below is a potential resolution provided by the Search Agent to resolve the user's request. You can use this resolution as a reference to help you in your task.

<potential_resolution>
{search_agent_solution}
</potential_resolution>
"""

assistant_prefill_message_for_code_suggestions = Prompt(
    """{% if context.use_case == "ci" %}
I'm implementing the proposed solution to fix the failing CI and address the root cause of the errors.
I will address all of the sub-requests. After each code_change I will make sure I've completed the task before moving onto the next one.
I will implement each sub-request in an individual code_change.
I will cite each change in the proposed solution before implementing it.

<user_response>
{% endif %}"""
)



import json


# function to iterate through a dictionary and ensure all values are json serializable
# truncates strings at 500 for sake of readability
def make_serializable(obj):
    MAX_STRING_LENGTH = 500

    if isinstance(obj, dict):
        new_dict = {}
        for key, value in obj.items():
            new_dict[key] = make_serializable(value)
        return new_dict
    elif isinstance(obj, list):
        return [make_serializable(item) for item in obj]
    else:
        try:
            json.dumps(obj)
            return obj
        except TypeError:
            try:
                stringified = str(obj)[:MAX_STRING_LENGTH]
                if len(stringified) == MAX_STRING_LENGTH:
                    stringified += "..."
                return stringified
            except Exception:
                return "Unserializable"


def make_serializable_dict(dictionary: dict):
    return make_serializable(dictionary)



from typing import Literal, Optional, TypedDict


class CheckStatus(TypedDict):
    message: str
    stdout: str
    succeeded: Optional[bool]
    status: Literal["pending", "running", "success", "failure", "cancelled"]
    llm_message: str
    container_name: str
    start_time: int
    end_time: int


# Status can be one of: completed, action_required, cancelled, failure, neutral, skipped, stale, success, timed_out, in_progress, queued, requested, waiting, pending.
gha_to_check_status = {
    "completed": "success",
    "action_required": "success",
    "cancelled": "cancelled",
    "failure": "failure",
    "neutral": "success",
    "skipped": "success",
    "stale": "success",
    "success": "success",
    "timed_out": "failure",
    "in_progress": "running",
    "queued": "pending",
    "requested": "pending",
    "waiting": "pending",
    "pending": "pending",
}

gha_to_succeeded = {
    "completed": True,
    "action_required": False,
    "cancelled": False,
    "failure": False,
    "neutral": True,
    "skipped": True,
    "stale": True,
    "success": True,
}

gha_to_message = {
    "completed": "Github Action completed",
    "action_required": "Github Action action required",
    "cancelled": "Github Action cancelled",
    "failure": "Github Action failed",
    "neutral": "Github Action neutral",
    "skipped": "Github Action skipped",
    "stale": "Github Action stale",
    "success": "Github Action succeeded",
    "timed_out": "Github Action timed out",
    "in_progress": "Github Action in progress",
    "queued": "Github Action queued",
    "requested": "Github Action requested",
    "waiting": "Github Action waiting",
    "pending": "Github Action pending",
}



import hashlib
import hmac

from sweepai.config.server import WEBHOOK_SECRET


def hash_sha256(text: str):
    return hashlib.sha256(text.encode("utf-8", "ignore")).hexdigest()


def verify_signature(payload_body: bytes, signature_header: str | None):
    """Verify that the payload was sent from GitHub by validating SHA256.

    Raise and return 403 if not authorized.

    Args:
        payload_body: original request body to verify (request.body())
        signature_header: header received from GitHub (x-hub-signature-256)
    """
    if not WEBHOOK_SECRET:
        # If the secret is not set, we can't verify the signature
        return True
    if not signature_header:
        return False
    hash_object = hmac.new(WEBHOOK_SECRET.encode("utf-8"), msg=payload_body, digestmod=hashlib.sha256)
    expected_signature = "sha256=" + hash_object.hexdigest()
    if not hmac.compare_digest(expected_signature, signature_header):
        return False
    return True



import os
import subprocess
from datetime import datetime, timedelta
from pathlib import Path

from apscheduler.schedulers.background import BackgroundScheduler
from loguru import logger
from tqdm import tqdm

from sweepai.core.github_utils import REPO_CACHE_BASE_DIR
from sweepai.utils.commit_metrics import get_commit_metrics
from sweepai.web.sync_clock import sync_time

scheduler = BackgroundScheduler()


def run_commit_metrics_computation():
    logger.info("Running commit metrics computation")
    access_token = os.environ.get("GITHUB_PAT")
    _ = get_commit_metrics(access_token, ignore_cache=True)
    # will be cached


def get_repo_commit_hash(repo_dir: str) -> str:
    return subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=repo_dir, text=True).strip()


def delete_repo_in_background(repo_dir: str):
    subprocess.Popen(
        ["bash", "-c", f"rm -rf {repo_dir} && echo 'Deleted repo {repo_dir}'"],
        start_new_session=True,
    )


def delete_old_repos():
    for base_repo_dir in tqdm(Path(REPO_CACHE_BASE_DIR).rglob("**/base")):
        logger.info(f"Checking {base_repo_dir} for old repos")
        base_dir = base_repo_dir.parent
        directories_to_remove = []
        for git_dir in base_dir.iterdir():
            if git_dir.name == "base" or len(git_dir.name) != 40 or git_dir.is_file():
                continue
            last_read_time = datetime.fromtimestamp(git_dir.stat().st_atime)
            directories_to_remove.append((git_dir, last_read_time))
        directories_to_remove.sort(key=lambda x: x[1])
        if directories_to_remove:
            directories_to_remove.pop()
        logger.info(f"Deleting {len(directories_to_remove)} old repos")
        for directory, last_read_time in directories_to_remove:
            if last_read_time < datetime.now() - timedelta(days=1):
                logger.info(f"Deleting {directory} (last read: {last_read_time})...")
                delete_repo_in_background(str(directory))


scheduler.add_job(
    run_commit_metrics_computation,
    "interval",
    seconds=60 * 60 * 12,  # 12 hours
    max_instances=1,
    next_run_time=datetime.now() + timedelta(minutes=30),
)

scheduler.add_job(sync_time, "interval", seconds=60 * 60, max_instances=1)  # once an hour
scheduler.add_job(delete_old_repos, "cron", hour=1, minute=0, max_instances=1)



import requests

from sweepai.config.server import LICENSE_KEY, NO_LICENSE
from sweepai.utils.cache import create_cache

cache = create_cache()


@cache.memoize(expire=60 * 5)  # 5 minute cache for failed requests
def fetch_license_data(license_key: str) -> requests.Response:
    return requests.post(
        "https://api.keygen.sh/v1/accounts/sweep-dev/licenses/actions/validate-key",
        headers={
            "Content-Type": "application/vnd.api+json",
            "Accept": "application/vnd.api+json",
        },
        json={"meta": {"key": license_key}},
    )


@cache.memoize(expire=60 * 60 * 24)  # 1 day cache for successful requests
def validate_license():
    # Validate the license key
    if NO_LICENSE:
        return True
    if LICENSE_KEY is None:
        raise ValueError(
            "No license key provided, please set the LICENSE_KEY environment variable or get a license key from https://deploy.sweep.dev."
        )
    response = fetch_license_data(LICENSE_KEY)
    if response.status_code != 200:
        raise ValueError(
            "License key is invalid or expired. Please contact us at team@sweep.dev to upgrade to an enterprise license."
        )
    obj = response.json()
    if obj["data"]["attributes"]["status"] not in ("ACTIVE", "EXPIRING"):
        raise ValueError(
            "License key is not active. Please contact us at team@sweep.dev to upgrade to an enterprise license."
        )
    return True


if __name__ == "__main__":
    assert validate_license()



num_of_snippets_to_query = 10
max_num_of_snippets = 5

INSTRUCTIONS_FOR_REVIEW = """\
> [!TIP]
> To get Sweep to edit this pull request, you can:
> * Comment below, and Sweep can edit the entire PR
> * Comment on a file, Sweep will only modify the commented file
> * Edit the original issue to get Sweep to recreate the PR from scratch"""



# IMPORTANT: to use the function decorator your function must have the username as the first param
import time
import traceback
from functools import wraps
from typing import Callable, TypeVar

from sweepai.o11y.event_logger import posthog
from sweepai.o11y.telemetry_utils import make_serializable_dict
from sweepai.utils.str_utils import get_hash
from sweepai.utils.streamable_functions import StreamableFunction

T = TypeVar("T")


def posthog_trace(
    function: Callable[..., T],
) -> Callable[..., T]:
    """
    Automatically posthog traces a function\n
    Requires username as the first parameter
    """
    if isinstance(function, StreamableFunction):
        raise ValueError(
            "posthog_trace cannot be used on a StreamableFunction, place @streamable before @posthog_trace"
        )

    @wraps(function)
    def wrapper(username: str, *args, metadata: dict = {}, **kwargs):
        tracking_id = get_hash()[:10]
        metadata = {**metadata, "tracking_id": tracking_id, "username": username}
        # attach args and kwargs to metadata
        if args:
            args_names = function.__code__.co_varnames[: function.__code__.co_argcount]
            args_dict = dict(zip(args_names[1:], args))  # skip first arg which must be username
            posthog_args: dict = make_serializable_dict(args_dict)
            metadata = {**metadata, **posthog_args}
        if kwargs:
            posthog_kwargs = make_serializable_dict(kwargs)
            if "access_token" in posthog_kwargs:
                del posthog_kwargs["access_token"]
            metadata = {**metadata, **posthog_kwargs}
        metadata = make_serializable_dict(metadata)
        posthog.capture(username, f"{function.__name__} start", properties=metadata)

        start_time = time.time()

        try:
            # check if metadata is in the function signature
            if "metadata" in function.__code__.co_varnames[: function.__code__.co_argcount]:
                result = function(username, *args, **kwargs, metadata=metadata)
            else:
                result = function(
                    username,
                    *args,
                    **kwargs,
                )
        except Exception as e:
            posthog.capture(
                username,
                f"{function.__name__} error",
                properties={
                    "error": str(e),
                    "trace": traceback.format_exc(),
                    "execution_time": time.time() - start_time,
                    **metadata,
                },
            )
            raise e
        else:
            posthog.capture(
                username,
                f"{function.__name__} success",
                properties={"execution_time": time.time() - start_time, **metadata},
            )
            return result

    return wrapper



import os

import psutil
from loguru import logger

from sweepai.backend.api_utils import get_cached_installation_id
from sweepai.core.github_utils import REPO_CACHE_BASE_DIR, ClonedRepo
from sweepai.search.agent.entity_search import EntitiesIndex
from sweepai.search.agent.summarize_directory import recursively_summarize_directory
from sweepai.search.query.ticket_utils import prep_snippets


def on_push(repo_full_name: str, tracking_id: str | None = None):
    try:
        cpu_percent = psutil.cpu_percent(interval=1)
        logger.info(f"CPU utilization: {cpu_percent}%")
        memory_info = psutil.virtual_memory()
        logger.info(f"Memory utilization: {memory_info.percent}%")
        if cpu_percent >= 50 or memory_info.percent >= 50:
            logger.info(
                f"Skipping update for {repo_full_name} due to high CPU or memory utilization: {cpu_percent}%/{memory_info.percent}%"
            )
            return
        logger.info(f"Updating cloned_repo for {repo_full_name}")
        org_name, repo = repo_full_name.split("/")
        previously_existing_repo_dirs = os.listdir(os.path.join(REPO_CACHE_BASE_DIR, repo_full_name))
        cloned_repo = ClonedRepo(
            repo_full_name, installation_id=get_cached_installation_id(org_name)
        )  # should have latest repo_dir
        is_cloned_repo_dir_the_latest_one = cloned_repo.repo_dir in previously_existing_repo_dirs
        if not is_cloned_repo_dir_the_latest_one:
            prep_snippets(
                cloned_repo=cloned_repo,
                query="DUMMY QUERY",
                skip_reranking=True,
            )
            _, _ = recursively_summarize_directory(cloned_repo, update_summaries=True)
            EntitiesIndex.from_dir(
                cloned_repo.repo_dir,
                key=f"{repo_full_name}:{cloned_repo.branch}",
                no_cache=True,
            )
    except Exception as e:
        logger.error(f"Error updating repo {repo_full_name}: {str(e)}")



import re
from dataclasses import dataclass, field

from numpy import array
from tqdm import tqdm

from sweepai.core.entities import Message
from sweepai.core.github_utils import ClonedRepo, MockClonedRepo
from sweepai.core.llm.chat import Thread

# from sweepai.logn.cache import file_cache
from sweepai.search.query.cohere_utils import cohere_rerank_call
from sweepai.search.query.vector_db import multi_get_query_texts_similarity

# TODO: implement into main agent

NUM_COMMITS = 10000

generate_query_instructions = """Objective: Assist an engineer in resolving a GitHub issue by crafting a precise natural language search query. This query will be used in a vector search engine to find solutions from similar past issues.
Steps:
- Identify the existing feature in the codebase that most closely resembles the current issue.
    - Example: If the current issue involves adding a filter to hide deleted posts, a similar feature might be adding a filter to hide deleted users.
- Compose a high-quality search query designed to locate the commit that implemented the identified similar feature.
- The search query should be specific and targeted, using relevant keywords and phrases to narrow down the search results effectively.
- Provide an analysis that outlines the reasoning behind the selected similar feature and the thought process used to create the search query.
- Present the final search query within XML blocks to ensure it is clearly distinguishable and easy to extract.

Format:
<analysis>
The COT block should include:
An explanation of why the identified feature is considered the most similar to the current issue.
A breakdown of the key components and considerations used to construct the search query.
Any assumptions made or additional information that could help refine the search query.
</analysis>

Final Search Query:
<query>
Insert the final natural language search query here, ensuring it is specific, targeted, and includes relevant keywords and phrases to effectively narrow down the search results. It should be a command starting with "Find the commit that...". Do NOT reference any files.
</query>

Note: The example provided below is for illustration purposes only. When using this prompt, focus on the actual input issue and codebase to identify the most relevant similar feature and craft an appropriate search query.

Example:

Current Issue: Implement a feature to allow users to mark comments as spoilers, hiding the content until clicked.

Chain of Thought (COT):
<cot>
The most similar existing feature to marking comments as spoilers is the implementation of the "Show/Hide" functionality for sensitive content in posts. This feature allows users to hide sensitive content behind a warning message, requiring a click to reveal the hidden content.

To search for the commit that implemented this feature, the query should reference specific entities such as the "Post" and "Content" models, as well as relevant actions like "hide," "show," "click," and "reveal." Additionally, mentioning the "warning message" and the "sensitive" nature of the content can help pinpoint the desired commit more accurately.
</cot>

Final Search Query:
<query>
Find the commit that introduced the functionality to hide sensitive content within a Post entity behind a warning message, allowing users to click and reveal the hidden Post.Content, similar to a "spoiler" feature for comments.
</query>"""

generate_query_user_prompt = (
    """<relevant_files>
{relevant_files}
</relevant_files>

<github_issue>
{github_issue}
</github_issue>

"""
    + generate_query_instructions
)

commit_selection_format_prompt = """Respond in the following format. First think step-by-step, then select any commits to show to the engineer. You don't want to overload the engineer with noise so only pick commits containing code that would be very similar to the final implementation.

<thinking>
Summarize each commit and the issue and think step-by-step. Then determine which commits contain code that would be very similar to the final implementation for this current feature, as well as which files from each commit contains relevant information. Avoid adding files with large diffs that are not relevant to the current feature.
</thinking>

<selected_commits>
<commit>
<sha>The commit sha.</sha>
<file_paths>
File paths, one per line, that contain relevant diffs.
</file_paths>
[additional file paths with relevant diffs]
</commit>
[additional commits, if any]
</selected_commits>"""

commit_selection_system_prompt = (
    """Your job is to help an engineer solve a GitHub issue by finding past solutions to similar issues. For example, if the current issue is that the app is showing deleted posts and previously a developer fixed a bug relating to the app showing deleted users, the implementations would be very similar and therefore this can be helpful.

You will list any previous commits that could be a potential reference implementation to the feature requested by the user.

"""
    + commit_selection_format_prompt
)

commit_selection_user_prompt = (
    """<github_issue>
{github_issue}
</github_issue>

<similar_commits>
{similar_commits}
</similar_commits>

"""
    + commit_selection_format_prompt
)


def get_diff_mapping(patch: str) -> dict[str, str]:
    """
    Returns file path to diff hunk mapping
    """
    diff_mapping = {}
    for hunk in patch.split("diff --git "):
        if not hunk.strip():
            continue
        file_path = hunk.split(" b/")[0].removeprefix("a/")
        diff_mapping[file_path] = "diff --git " + hunk
    return diff_mapping


@dataclass
class Commit:
    sha: str
    message: str
    diff: str = field(repr=False)

    @property
    def xml(self):
        return f"<commit>\n<commit_sha>{self.sha}</commit_sha>\n<commit_message>\n{self.message}\n</commit_message>\n<source>\n{self.diff}\n</source>\n</commit>"


def generate_query(query: str, cloned_repo: ClonedRepo, relevant_file_path: list[str]) -> str:
    thread = Thread(
        messages=[
            Message(role="system", content=generate_query_instructions),
        ]
    )
    relevant_files_string = ""
    file_contents = []
    for file_path in relevant_file_path:
        try:
            file_contents = cloned_repo.get_file_contents(file_path)
        except FileNotFoundError:
            continue
        relevant_files_string += f"<relevant_file>\n<file_path>\n{file_path}\n</file_path>\n<source>\n{file_contents}\n</source>\n</relevant_file>"
    relevant_files_string = f"<relevant_files>\n{relevant_files_string}\n</relevant_files>"
    user_message = generate_query_user_prompt.format(relevant_files=relevant_files_string, github_issue=query)
    response = thread.chat(
        user_message,
        model="claude-3-opus-20240229",  # Sonnet doesn't work well with this prompt
    )
    match_ = re.search(r"<query>(.*?)</query>", response, re.DOTALL)
    if match_:
        return match_.group(1).strip()
    else:
        return query


def query_relevant_commits(query: str, cloned_repo: ClonedRepo, relevant_file_paths: list[str]) -> list[Commit]:
    # only get git blames of the file
    last_commits = []
    viewed_commits = set()
    all_lines = []
    for file_path in relevant_file_paths:
        all_lines.extend(cloned_repo.git_repo.git.blame("HEAD", "--", file_path).splitlines())

    for line in tqdm(all_lines):
        if not line:
            continue
        sha, _message = line.split(" ", 1)
        if sha in viewed_commits:
            continue
        viewed_commits.add(sha)
        diff = cloned_repo.git_repo.git.diff(f"{sha}~1", sha)
        full_message = cloned_repo.git_repo.git.show("--no-indent", "--no-patch", "--format=%s%n%n%b", sha)
        last_commits.append(Commit(sha=sha, message=full_message, diff=diff))

    enhanced_query = generate_query(query, cloned_repo, relevant_file_paths)
    similarities = array(
        multi_get_query_texts_similarity(
            [enhanced_query],
            [last_commit.message + last_commit.diff for last_commit in last_commits],
        )
    )
    top_indices = [index for index in similarities.flatten().argsort()[-1000:][::-1]]
    top_documents = [last_commits[index].message + last_commits[index].diff for index in top_indices]
    embed_index_to_rerank_index = {rank_index: actual_index for rank_index, actual_index in enumerate(top_indices)}

    NUM_DIFFS = 5
    try:
        response = cohere_rerank_call(enhanced_query, top_documents)
        commits = []
        for document in response.results[:NUM_DIFFS]:
            index = embed_index_to_rerank_index[document.index]
            commit = last_commits[index]
            commits.append(commit)
    except Exception as e:
        print(e)
        commits = [last_commits[index] for index in top_indices[:NUM_DIFFS]]

    return commits


def get_relevant_commits(query: str, cloned_repo: ClonedRepo, relevant_file_paths: list[str]) -> list[Commit]:
    commits = query_relevant_commits(query, cloned_repo, relevant_file_paths)
    commits_str = "\n".join([commit.xml for commit in commits])
    user_prompt = commit_selection_user_prompt.format(github_issue=query, similar_commits=commits_str)
    thread = Thread(
        messages=[
            Message(role="system", content=commit_selection_system_prompt),
        ]
    )
    response = thread.chat(
        user_prompt,
        model="claude-3-opus-20240229",  # Sonnet fails at this task
    )
    commits_pattern = re.compile(
        r"<commit>\s*?<sha>(?P<sha>.*?)</sha>\s*?<file_paths>\s*?(?P<file_paths>.*?)\s*?</file_paths>\s*?</commit>",
        re.DOTALL,
    )
    result_str = ""
    for match in commits_pattern.finditer(response):
        commit_sha = match.group("sha").strip()
        commit = next(commit for commit in commits if commit.sha == commit_sha)
        file_paths = match.group("file_paths").strip().splitlines()
        diff_mapping = get_diff_mapping(commit.diff)
        filtered_diffs = "\n".join([diff_mapping[file_path] for file_path in file_paths if file_path in diff_mapping])
        result_str += f"<commit>\n<sha>{commit_sha}</sha>\n<commit_message>\n{commit.message}\n</commit_message>\n<source>\n{filtered_diffs}\n</source>\n</commit>\n"
    result_str = f"<relevant_existing_commits>\n{result_str.strip()}\n</relevant_existing_commits>"
    return result_str


if __name__ == "__main__":
    from sweepai.core.github_utils import MockClonedRepo

    query = """Where is the code that handles the API?"""
    directory = "/mnt/sweep_benchmark/sweep"
    relevant_file_paths = ["sweepai/api.py"]
    cloned_repo = MockClonedRepo(_repo_dir=directory, repo_full_name="sweepai/sweep")
    print(get_relevant_commits(query, cloned_repo, relevant_file_paths))



import time
from queue import Empty, Queue
from threading import Thread
from typing import Iterator

from fastapi.responses import StreamingResponse

NOT_SET = object()


def StreamingResponseWithHeartbeat(
    stream_response: Iterator,
    heartbeat_interval: float = 3,
    timeout: float = 30 * 60,
    heartbeat_message: str = NOT_SET,
    **kwargs,
):
    return StreamingResponse(
        stream_with_heartbeat(stream_response, heartbeat_interval, timeout, heartbeat_message, **kwargs)
    )


def stream_with_heartbeat(
    iterator: Iterator[str],
    heartbeat_interval: float = 3,
    timeout: float = 30 * 60,
    heartbeat_message: str = NOT_SET,
):
    queue = Queue()

    def push_thread():
        for item in iterator:
            queue.put(item)

    thread = Thread(target=push_thread)
    thread.start()
    start_time = time.time()
    last_item = None
    while thread.is_alive() and time.time() - start_time < timeout:
        try:
            last_item = queue.get(timeout=heartbeat_interval)
            yield last_item
        except Empty:
            if heartbeat_message is not NOT_SET:
                yield heartbeat_message
            else:
                yield last_item
    if not thread.is_alive():
        # Yield the rest of the items
        while not queue.empty():
            last_item = queue.get()
            yield last_item
    thread.join()



"""
This script is used to compute metrics for merged code suggestions. It is used to determine the percentage of lines added in a PR that were suggested by Sweep.
"""

import json
import os
import re
from dataclasses import dataclass
from datetime import datetime
from random import randint

from diskcache import Cache
from github.Repository import Repository
from loguru import logger

from sweepai.backend.api_utils import (
    get_cached_installation_id,
    get_github_client_from_org,
)
from sweepai.config.server import CACHE_DIRECTORY
from sweepai.core.github_utils import ClonedRepo
from sweepai.utils.diff import generate_diff
from sweepai.utils.str_utils import strip_triple_quotes

metrics_cache = Cache(f"{CACHE_DIRECTORY}/metrics_cache")

TWENTY_FOUR_HOURS = 24 * 60 * 60
SIX_HOURS = 6 * 60 * 60
CACHE_VERSION = "0.0.1"
EMPTY_METRIC_PLACEHOLDER = "EMPTY_CACHE_VALUE"


# can be reworked later for actual display, this will do for now
def get_highlighted_lines(lines: list[str], common_lines: list[str]) -> str:
    formatted_lines = []
    for line in lines:
        if line in common_lines:
            formatted_lines.append(f"\033[94m{line}\033[0m")  # Blue color
        else:
            formatted_lines.append(line)
    return "\n".join(formatted_lines)


def get_pr_for_branch(repo: Repository, org_name: str, branch: str):
    pull_requests = repo.get_pulls(state="closed", head=f"{org_name}:{branch}")
    max_prs = 2
    for idx, pr in enumerate(pull_requests):
        if idx >= max_prs:
            break
        logger.info(pr.head.ref)
        if pr.head.ref == branch and pr.merged:
            return pr.html_url, {file.filename: file.patch for file in pr.get_files()}
    return None, {}


def get_sweep_code_suggestions(all_messages_joined):
    pattern = r"<code_change>\s*<file_path>\n*(?P<filePath>[\s\S]+?)\n*</file_path>\s*(<justification>\n*(?P<justification>[\s\S]+?)\n*</justification>)?\s*(<original_code>\n*(?P<originalCode>[\s\S]*?)</original_code>\n*)?<new_code>\n*(?P<newCode>[\s\S]+?)\n*(</new_code>)?\s*</code_change>"
    sweep_code_suggestions = [
        {
            "filePath": match_.groupdict().get("filePath", ""),
            "originalCode": strip_triple_quotes(match_.groupdict().get("originalCode") or ""),
            "newCode": strip_triple_quotes(match_.groupdict().get("newCode") or ""),
        }
        for match_ in re.finditer(pattern, all_messages_joined, re.DOTALL)
    ]
    return sweep_code_suggestions


def extract_changed_lines(diff):
    changed_lines = []
    for line in diff.split("\n"):
        # added line
        if line.startswith("+") and not line.startswith("+++"):
            changed_lines.append(line[1:])
        # removed line
        elif line.startswith("-") and not line.startswith("---"):
            changed_lines.append(line[1:])
    return changed_lines


def find_common_lines(changes_dict: dict[str, list[str]], second_list: list[str]) -> list[str]:
    stripped_second_list = [line.strip() for line in second_list]
    common_lines = []
    for _, changed_lines in changes_dict.items():
        for line in changed_lines:
            if line.strip() in stripped_second_list:
                common_lines.append(line)
    return common_lines


@dataclass
class Metrics:
    percentage_of_lines_added: float
    number_of_lines_added: int
    highlighted_lines: str
    common_added_lines: list[str]
    pull_request_url: str
    message_id: str
    last_modified_date: str
    org_name: str
    repo_name: str
    username: str


def compute_metrics(
    sweep_code_suggestions,
    file_patches: dict[str, str],
    metadata: dict,
    latest_cloned_repos: dict[str, ClonedRepo] = {},
) -> Metrics:
    generated_diffs = [
        (
            suggestion["filePath"],
            generate_diff(suggestion["originalCode"], suggestion["newCode"], n=0),
        )
        for suggestion in sweep_code_suggestions
    ]
    sweep_suggested_changed_lines = {}
    for file_path, diff in generated_diffs:
        sweep_suggested_changed_lines[file_path] = extract_changed_lines(diff)

    pr_changed_lines = []
    common_added_lines = []
    # this checks if the PR already exists
    if metadata["did_pr_exist"]:
        for patch in file_patches.values():
            pr_changed_lines.extend(extract_changed_lines(patch))
        common_added_lines = find_common_lines(sweep_suggested_changed_lines, pr_changed_lines)
    else:
        # check the file in the cloned repo and then see if the sweep's suggestion exists in the file
        cloned_repo = get_cloned_repo(
            repo_full_name=metadata["repo_full_name"],
            latest_cloned_repos=latest_cloned_repos,
        )
        for file_path, changed_lines in sweep_suggested_changed_lines.items():
            try:
                file_contents = cloned_repo.get_file_contents(file_path)
                for line in changed_lines:
                    if line in file_contents:
                        common_added_lines.append(line)
            except FileNotFoundError:
                logger.warning(f"File not found (likely renamed):")
    number_of_lines_added = len(common_added_lines)
    total_pr_lines = max(len(pr_changed_lines), 1)  # Avoid division by zero
    percentage_of_lines_added = 0 if total_pr_lines <= 1 else int((number_of_lines_added / total_pr_lines) * 100)
    pull_request_url = metadata["pull_request_url"] if metadata["did_pr_exist"] else "N/A"

    org_name, repo_name = metadata["repo_full_name"].split("/")

    return Metrics(
        percentage_of_lines_added=percentage_of_lines_added,
        number_of_lines_added=number_of_lines_added,
        highlighted_lines=get_highlighted_lines(pr_changed_lines, common_added_lines),
        common_added_lines=common_added_lines,
        pull_request_url=pull_request_url,
        message_id=metadata["message_id"],
        last_modified_date=metadata["last_modified_date"],
        org_name=org_name,
        repo_name=repo_name,
        username=metadata["username"],
    )


def get_data_from_json_path(json_path: str):
    data = {}
    try:
        # get the last time the file was modified
        with open(json_path, "r") as file:
            data = json.load(file)
        last_modified = os.path.getmtime(json_path)
        data["last_modified"] = last_modified
    except Exception as e:
        logger.error(f"Error loading json file: {e}")
    return data


def get_pr_and_file_patches(data: dict) -> tuple[str, dict[str, str], bool]:
    branch = data["branch"]
    repo_full_name = data["repo_name"]
    pr_cache_key = f"pr-{repo_full_name}-{branch}-v{CACHE_VERSION}"
    pr_cache_hit = metrics_cache.get(pr_cache_key)
    if pr_cache_hit:
        return pr_cache_hit
    org_name = repo_full_name.split("/")[0] if "/" in repo_full_name else None
    if not org_name:
        raise ValueError(f"Could not determine org name from repo name: {repo_full_name}")
    try:
        _, github_client = get_github_client_from_org(org_name)
    except Exception:
        raise Exception(f"Error getting github client for {repo_full_name}")
    repo = github_client.get_repo(repo_full_name)
    pull_request_url, file_patches = get_pr_for_branch(repo=repo, org_name=org_name, branch=branch)
    if pull_request_url:
        metrics_cache.set(
            pr_cache_key, (pull_request_url, file_patches, True)
        )  # if a pr exists, cache it with no expiration
    did_pr_exist = pull_request_url is not None
    return pull_request_url, file_patches, did_pr_exist


def get_metrics_for_json_path(json_path: str, latest_cloned_repos: dict[str, ClonedRepo] = {}):
    """This file is the topmost cache layer for metrics. It will check if the metrics are already cached and return them if they are. If not, it will compute the metrics and cache them."""
    cache_key = f"metrics-{json_path}-v{CACHE_VERSION}"
    metrics_cache_hit = metrics_cache.get(cache_key)
    if metrics_cache_hit and metrics_cache_hit != EMPTY_METRIC_PLACEHOLDER:
        return metrics_cache_hit
    try:
        data = get_data_from_json_path(json_path=json_path)
        all_messages_joined = "\n\n".join([message["content"] for message in data["messages"]])
        sweep_code_suggestions = get_sweep_code_suggestions(all_messages_joined)
        pull_request_url, file_patches, did_pr_exist = get_pr_and_file_patches(data)
        pull_request_url = pull_request_url or "N/A"
        message_id = ""
        if json_path and "/" in json_path:
            message_id = json_path.split("/")[-1].replace(".json", "")
        metadata = {
            "pull_request_url": pull_request_url,
            "message_id": message_id,
            "last_modified_date": datetime.fromtimestamp(data["last_modified"]).strftime("%Y-%m-%d %H:%M:%S"),
            "repo_full_name": data["repo_name"],
            "did_pr_exist": did_pr_exist,
            "username": data["username"],
        }
        # all computation happens in compute_metrics
        computed_metrics = compute_metrics(
            sweep_code_suggestions=sweep_code_suggestions,
            file_patches=file_patches,
            metadata=metadata,
            latest_cloned_repos=latest_cloned_repos,
        )
        if computed_metrics.number_of_lines_added > 0 or did_pr_exist:
            metrics_cache.set(cache_key, computed_metrics)  # This does not expire
        return computed_metrics
    except Exception as e:
        logger.warning(f"Error computing metrics for {json_path}: {e}")
        metrics_cache.set(
            cache_key, EMPTY_METRIC_PLACEHOLDER, expire=SIX_HOURS * randint(1, 5)
        )  # add jitter to avoid cache stampede
        return EMPTY_METRIC_PLACEHOLDER


def get_cloned_repo(repo_full_name: str, latest_cloned_repos: dict[str, ClonedRepo]):
    if repo_full_name in latest_cloned_repos:
        return latest_cloned_repos[repo_full_name]
    org_name = repo_full_name.split("/")[0]
    try:
        installation_id = get_cached_installation_id(org_name)
    except Exception as e:
        raise Exception(f"Error getting installation id for {org_name}: {e}")
    cloned_repo = ClonedRepo(repo_full_name=repo_full_name, installation_id=installation_id)
    latest_cloned_repos[repo_full_name] = cloned_repo
    return cloned_repo


def compute_metrics_for_all_messages(csv_file: str):
    logger.info(f"Computing metrics for all messages")
    # sort the jsons by last updated first
    all_message_jsons = [
        (
            f"{CACHE_DIRECTORY}/messages/{file}",
            os.path.getmtime(f"{CACHE_DIRECTORY}/messages/{file}"),
        )
        for file in os.listdir(f"{CACHE_DIRECTORY}/messages")
        if file.endswith(".json")
    ]
    all_message_jsons = [file for file, _ in sorted(all_message_jsons, key=lambda x: x[1], reverse=True)]
    # make this sequential to be less costly + to allow safe sharing of cloned repos dictionary
    latest_cloned_repos = {}
    all_metrics = []
    for json_path in all_message_jsons:
        metrics = get_metrics_for_json_path(json_path, latest_cloned_repos)
        if metrics != EMPTY_METRIC_PLACEHOLDER and metrics is not None:
            all_metrics.append(
                {
                    "last_modified_date": metrics.last_modified_date,
                    "message_link": metrics.message_id,
                    "number_of_pr_lines_merged": metrics.number_of_lines_added,
                    "percentage_of_pr_lines_merged": metrics.percentage_of_lines_added,
                    "pull_request_url": metrics.pull_request_url,
                    "org_name": metrics.org_name,
                    "repo_name": metrics.repo_name,
                    "username": metrics.username,
                }
            )
    total_messages = len(all_message_jsons)
    successful_messages_counted = len(all_metrics)

    # Calculate sessions per day
    sessions_per_day = {}
    for metric in all_metrics:
        date = metric["last_modified_date"].split()[0]  # Extract date part
        sessions_per_day[date] = sessions_per_day.get(date, 0) + 1

    # Update all_metrics with sessions count; we turned off message forking so this is 1:1
    for metric in all_metrics:
        date = metric["last_modified_date"].split()[0]
        metric["sessions_count"] = sessions_per_day[date]
    from pandas import DataFrame

    df = DataFrame(all_metrics)
    # Dedupe rows, keeping the one with the greatest number of lines changed for a given PR URL
    df = df.sort_values("number_of_pr_lines_merged", ascending=False).drop_duplicates(
        subset="pull_request_url", keep="first"
    )
    df.sort_values(by="last_modified_date", inplace=True)
    df.to_csv(csv_file, index=False, na_rep="N/A")
    logger.info(
        f"Metrics written to {csv_file}\n{successful_messages_counted}/{total_messages} messages, {round((successful_messages_counted / total_messages) * 100, 1)}% successfully computed"
    )
    return all_metrics


if __name__ == "__main__":
    csv_file = f"{CACHE_DIRECTORY}/metrics.csv"
    all_metrics = compute_metrics_for_all_messages(csv_file=csv_file)



from sweepai.core.llm.chat import Thread, continuous_llm_calls
from sweepai.utils.format_utils import Prompt

system_message = "You are a helpful assistant that will help determine if a set of files is sufficient for answering a technical question about a codebase, or if additional searching in the codebase is required."

user_message_prompt = Prompt(
    """
Here are your relevant files:
<relevant_files>
{{ snippets|render_snippets }}
</relevant_files>

Here is the user's question:
<user_request>
{{ question }}
</user_request>

Respond in the following XML format:
<analysis>

Given a set of code snippets and a question about a codebase, analyze the snippets to determine if there is sufficient information to answer the question. Provide your analysis in the following XML format:

<format>
<analysis>
a. Break down the user's question to determine what information you need.
b. For each provided snippet that is relevant to the question:
    1. [List each relevant code symbol defined in this snippet. Then, for each symbol, give an extremely detailed explanation of what information it contains. If it is a complex function or class, give an extremely detailed step by step explanation of how it works.]
</analysis>
<relevant_snippets>
<snippet>path/to/file.py:1-10</snippet>
<snippet>path/to/another/file.py:12-20</snippet>
<!-- Add more snippet tags as needed -->
</relevant_snippets>
<missing_information>
<item>[Describe a piece of missing information]</item>
<item>[Describe another piece of missing information]</item>
<!-- Add more item tags as needed -->
</missing_information>
<conclusion>
[For each missing piece of information, think critically about whether it is necessary to answer the question or present in the snippets.]
</conclusion>
<sufficient_information>[true/false]</sufficient_information>
<confidence_level>[0-100]</confidence_level>
</format>
"""
)


def use_instant_reply(question: str, snippets: list[str]):
    thread = Thread.from_system_message_string(system_message)
    user_message = user_message_prompt.render(question=question, snippets=snippets)

    response = continuous_llm_calls(
        username="anonymous",
        thread=thread,
        content=user_message,
        model="claude-3-opus-20240229",  # Sonnet has context issues
    )

    return "<sufficient_information>true</sufficient_information>" in response



import time
from dataclasses import dataclass

from loguru import logger


@dataclass
class Timer:
    name: str = ""
    min_time: float = 0.01
    start: float = 0
    end: float = 0
    time_elapsed: float = -1
    do_print: bool = True
    precision: int = 2

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end = time.time()
        self.time_elapsed = self.end - self.start
        if self.do_print and self.time_elapsed > self.min_time:
            if self.name:
                logger.debug(f"Timer {self.name} elapsed: {self.time_elapsed:.{self.precision}f} seconds")
            else:
                logger.debug(f"Timer elapsed: {self.time_elapsed:.{self.precision}f} seconds")


if __name__ == "__main__":
    with Timer() as t:
        time.sleep(1)
    print(t.time_elapsed)
    assert t.time_elapsed > 0.9



from github.Commit import Commit
from github.GithubException import GithubException
from github.PullRequest import PullRequest
from loguru import logger
from tqdm import tqdm

from sweepai.backend.api import fix_issue
from sweepai.backend.api_utils import get_pr_snippets
from sweepai.config.server import (
    FIX_CI_ALLOWLIST,
    GITHUB_BASE_URL,
)
from sweepai.core.entities import Snippet, snippet_to_string, organize_snippets
from sweepai.core.github_utils import ClonedRepo, get_github_repo, get_installation_id
from sweepai.dataclasses.chat_agent_context import ChatAgentContext
from sweepai.dataclasses.fix_issue_result import FixIssueResult
from sweepai.handlers import handlers_cache
from sweepai.handlers.format_issue_results import (
    extract_diff_patches,
    format_results,
    format_results_body,
)
from sweepai.handlers.on_comment import (
    generate_pull_request_metadata,
    generate_revert_comment_and_format_results,
    get_actor_from_pull_request,
)
from sweepai.ci.cleanup_ci import clean_ci_logs
from sweepai.ci.summarize_history import summarize_history
from sweepai.modify.modify_utils import english_join
from sweepai.o11y.log_utils import log_to_file
from sweepai.search.agent.agent_utils import get_all_snippets_from_query
from sweepai.utils.format_utils import Prompt
from sweepai.utils.str_utils import (
    code_block,
    markdown_list,
    pack_items_for_prompt,
    tail,
    truncate_text_width,
    wildcard_match,
)
from sweepai.utils.ticket_rendering_utils import (
    HR,
    IssueErrorHandler,
    find_sweeps_github_comment,
    get_suffix_for_sweep_comment,
)
from sweepai.ci.fetch_ci import FixCIInput, create_hyperlink, get_ci_failure_logs, interval_timer, wait_for_ci_failures

### CACHE LOGIC

def get_existing_snippets(pull_request: PullRequest, cloned_repo: ClonedRepo) -> list[Snippet]:
    fix_issue_result: FixIssueResult | None = handlers_cache.get(pull_request.html_url, None)
    if not isinstance(fix_issue_result, FixIssueResult):
        return []
    snippets = fix_issue_result.snippets # from on_ticket.py
    if fix_ci_history := get_ci_history(pull_request):
        # TODO: add logic to clear reverted changes
        _, last_result = fix_ci_history[-1]
        if not last_result:
            return []
        snippets = last_result.snippets
    # replace/handle contents and lines here because they might be stale from a previous commit
    cleaned_and_deduped_snippets = []
    for snippet in snippets:
        try:
            snippet.content = cloned_repo.get_file_contents(snippet.file_path)
            snippet.end = min(snippet.end, snippet.content.count("\n"))
            cleaned_and_deduped_snippets.append(snippet)
        except FileNotFoundError:
            logger.info(f"File {snippet.file_path} not found in cloned repo")
            continue
    # deduplicate snippets
    cleaned_and_deduped_snippets = organize_snippets(cleaned_and_deduped_snippets)
    return cleaned_and_deduped_snippets


def get_ci_history(
    pull_request: PullRequest,
) -> list[tuple[FixCIInput, FixIssueResult]]:
    repo = pull_request.base.repo
    handlers_cache_key = f"ci_fix_history-{pull_request.html_url}"
    history = handlers_cache.get(handlers_cache_key, [])
    if history and isinstance(history[0], dict):  # for temporary backwards compatibility
        history = [FixIssueResult(**result) for result in history]
    return [
        (inp, result)
        for inp, result in history
        if "revert" not in repo.get_commit(inp.commit_hash).commit.message.lower()
    ]


def append_to_ci_history(pull_request: PullRequest, history: list[tuple[FixCIInput, FixIssueResult]]):
    handlers_cache_key = f"ci_fix_history-{pull_request.html_url}"
    handlers_cache.set(handlers_cache_key, [*handlers_cache.get(handlers_cache_key, []), *history])


# -%} will make it single line joined
render_history_prompt = Prompt(
    """
{% for inp, result in history -%}
<fix_attempt index="{{ loop.index }}">
{{ inp.cleaned_logs | trim('\n') | wrap_xml("error_logs") }}
{{ format_results_body(result) | trim('\n') | wrap_xml("fixes") }}
</fix_attempt>
{% endfor %}
""",
    globals=[format_results_body],
)


def render_history(pull_request: PullRequest) -> str:
    history = get_ci_history(pull_request)
    # sometimes we write bad PRs to the cache
    history = [(inp, result) for inp, result in history if result.pull_request_data]
    history_str = render_history_prompt.render(history=history)
    return history_str


def get_last_error_log(cleaned_logs: str) -> str:
    lines = cleaned_logs.splitlines()
    MAX_LINE_LENGTH = 300
    backup_line = lines[-1][:MAX_LINE_LENGTH]
    lines = [line for line in lines if len(line) < MAX_LINE_LENGTH]
    if answer := next((line for line in lines[::-1] if "Error" in line), None):
        return answer
    if answer := next((line for line in lines[::-1] if "error" in line.lower()), None):
        return answer
    return backup_line


render_history_for_github_comment_prompt = Prompt(
    """
{% if history and len(history) > 0 %}
I'm currently working on fixing the CI errors. Here were all my previous attempts:

{% if len(history) > 10 %}
...
{% endif %}

{% for (inp, result), commit_message in zip(history, commit_messages) -%}
{% if loop.index > len(history) - 10 %}
<details>
<summary>{{ loop.index }}. <a href="https://{{GITHUB_BASE_URL}}/{{repo.full_name}}/commit/{{inp.commit_hash}}">{{commit_message}} ({{inp.commit_hash[:7]}})</a> to fix the following error:

<pre>
{{ inp.cleaned_logs | tail(max_lines=20) | truncate_text_width(width=200) }}
</pre>
</summary>

#### Attempted Fix

<blockquote>

{{ extract_diff_patches(result) }}
</blockquote>
</details>
{% endif %}
{% endfor %}

---

{% endif %}
""",
    globals=[extract_diff_patches, zip, get_last_error_log, len],
    filters=[tail, english_join, truncate_text_width],
    constants={"GITHUB_BASE_URL": GITHUB_BASE_URL},
)


def get_patch_from_commit(commit: Commit) -> str:
    return "\n\n".join(data["patch"] for data in commit.raw_data["files"])


def render_history_for_github_comment(
    pull_request: PullRequest, history: list[tuple[FixCIInput, FixIssueResult]]
) -> str:
    # TODO: add summary of number of lines changed
    # TODO: add timestamps
    repo = pull_request.base.repo
    commit_data = {}
    children = {}
    pr_commits = list(tqdm(pull_request.get_commits(), desc="Fetching commits"))
    for commit, next_commit in zip(pr_commits, [None] + pr_commits):
        commit_data[commit.sha] = commit
        if next_commit:
            children[next_commit.sha] = commit.sha
    commits = []
    commit_messages = []
    for inp, _result in history:
        # I don't fully trust the code here, very easy to get a key error, so I'm doing this
        if inp.commit_hash not in commit_data:
            logger.warning(f"Commit {inp.commit_hash} not found in commit_data or children")
            continue
        if inp.commit_hash not in children:
            logger.warning(f"Commit {inp.commit_hash} not found in children")
            continue
        if children[inp.commit_hash] not in commit_data:
            logger.warning(f"Commit {children[inp.commit_hash]} not found in commit_data")
            continue
        commits.append(commit_data[inp.commit_hash])
        commit_messages.append(commit_data[children[inp.commit_hash]].commit.message)
    return render_history_for_github_comment_prompt.render(history=history, repo=repo, commit_messages=commit_messages)


### MAIN


async def fix_ci_failures_once(
    pull_request: PullRequest,
    tracking_id: str | None = None,
    force: bool = False,
):
    if pull_request.state != "open" and not force:
        return
    comment = find_sweeps_github_comment(pull_request)
    with log_to_file(tracking_id), IssueErrorHandler(
        pull_request.as_issue(),
        error_title="Error while fixing CI",
        tracking_id=tracking_id,
    ) as monitor:
        logger.info(f"URL: {pull_request.html_url}")
        repo = pull_request.base.repo
        repo_full_name = repo.full_name
        org_name, repo_name = repo_full_name.split("/")
        access_token = repo._requester.auth.token
        installation_id = get_installation_id(username=org_name)
        username = get_actor_from_pull_request(pull_request)
        branch = pull_request.head.ref
        commit: Commit = pull_request.get_commits().reversed[0]
        commit_link = create_hyperlink(commit)
        SWEEP_SETTINGS_PAGE = f"https://{GITHUB_BASE_URL}/organizations/{org_name}/settings/installations/{installation_id}"  # TODO: support user configs

        ci_history = get_ci_history(pull_request)
        markdown_history = render_history_for_github_comment(pull_request, ci_history)
        message = (
            markdown_history
            + f"\n\nI'm currently monitoring {commit_link} for CI failures. This should take about 5-10 min."
            + get_suffix_for_sweep_comment(tracking_id)
        )
        if comment:
            comment.edit(body=message)
        else:
            comment = pull_request.create_issue_comment(message)
        monitor.comment = comment

        # Wait for CI failures
        try:
            pending_runs, successful_runs, error_runs = wait_for_ci_failures(pull_request)
        except GithubException as e:
            if e.status == 403:
                comment.edit(
                    body=f"I couldn't access the CI status. Please make sure to give Sweep access to your GitHub repository at https://{GITHUB_BASE_URL}/organizations/{org_name}/settings/installations/{installation_id}."
                    + get_suffix_for_sweep_comment(tracking_id)
                )
                return
            raise

        # Fetching logs
        if not error_runs:
            logger.info("No error runs found")
            if ci_history:
                # Used to have error runs, no longer do, meaning the CI errors have been fixed
                comment.edit(
                    body=markdown_history.replace(
                        "I'm currently working on fixing the CI errors. Here were all my previous attempts:",
                        "**✨ The CI errors have been fixed! ✨**\n\nHere's what I did:",
                    )
                    + f"\n\nIf you think this is an error, check Sweep's permissions at {SWEEP_SETTINGS_PAGE}."
                    + get_suffix_for_sweep_comment(tracking_id)
                )
            else:
                if pending_runs or successful_runs:
                    # Only pending or successful runs
                    comment.edit(
                        body=markdown_history
                        + f"\n\nNo CI failures were found. If you think this is an error, check Sweep's permissions at {SWEEP_SETTINGS_PAGE}."
                        + get_suffix_for_sweep_comment(tracking_id)
                    )
                else:
                    # No runs at all
                    comment.edit(
                        body=markdown_history
                        + f"\n\nNo CI runs were found. Sweep can fix CI failures, so we recommend setting it up for your repository to let Sweep run your build and tests.\n\nIf you think this is an error, check Sweep's permissions at {SWEEP_SETTINGS_PAGE}."
                        + get_suffix_for_sweep_comment(tracking_id)
                    )
            return
        logs_list = []
        # TODO: make this it's own function
        for i in interval_timer(total_time=60 * 10):
            logs_list = get_ci_failure_logs(installation_id=installation_id, runs=error_runs, pull_request=pull_request)
            if logs_list:
                logger.info(f"Found info logs here:\n {logs_list[-10000:]}")
                break

        if not logs_list:
            comment.edit(
                body=markdown_history
                + f"\n\nSome of the CI runs failed, but I couldn't access the logs. Check Sweep's permissions at {SWEEP_SETTINGS_PAGE}."
                + get_suffix_for_sweep_comment(tracking_id)
            )
            return

        if not wildcard_match(username, FIX_CI_ALLOWLIST):
            comment.edit(
                body=f"Hey @{username}, CI auto-fix is currently not enabled for you. Please reach out to the [Sweep team](william@sweep.dev) to enable it."
                + get_suffix_for_sweep_comment(tracking_id)
            )
            return

        joined_logs = markdown_list([create_hyperlink(run) for run in error_runs])
        initial_message = (
            markdown_history
            + f"\n\nI currently see that the following CI runs are failing for {commit_link}:\n{joined_logs}\n\nHere are some of the logs:\n\n```\n...\n{tail(logs_list[0], max_lines=20)}\n```\n\nI'm currently trying to fix them. This should take about 5-10 min."
        )
        comment.edit(body=initial_message + get_suffix_for_sweep_comment(tracking_id))
        cloned_repo = ClonedRepo.from_repo_full_name(repo_full_name=repo_full_name, branch=branch)

        pull_request_metadata = generate_pull_request_metadata(pull_request)
        _, _, pulls_message = get_pr_snippets(
            repo_name=repo_full_name,
            annotations={"pulls": pull_request_metadata},
            cloned_repo=cloned_repo,
            context=ChatAgentContext(use_case="ci"),
        )
        
        # Logs cleanup
        cleaned_logs = clean_ci_logs(
            username=username,
            error_logs=logs_list[0],
            pulls_message=pulls_message,
        )

        # Summarize history of previous CI runs
        summarized_history = summarize_history(username=username, history=render_history(pull_request), current_failing_error_logs=cleaned_logs)

        context = ChatAgentContext(
            use_case="ci",
            history=summarized_history,
        )

        initial_message = (
            markdown_history
            + f"\n\nI currently see that the following CI runs are failing for {commit_link}:\n{joined_logs}\n\nHere are some of the logs:\n\n{code_block(tail(cleaned_logs, max_lines=20))}\n\nI'm currently trying to fix them. This should take about 5-10 min."
        )
        comment.edit(body=initial_message + get_suffix_for_sweep_comment(tracking_id))

        query = cleaned_logs

        pull_request_changed_snippets = get_existing_snippets(pull_request=pull_request, cloned_repo=cloned_repo)
        ci_mentioned_snippets = get_all_snippets_from_query(query=cleaned_logs, cloned_repo=cloned_repo)
        # Put pull_request_changed_snippets second because it's at least in the diff
        ci_snippets: list[Snippet] = organize_snippets(ci_mentioned_snippets + pull_request_changed_snippets)
        # need to make sure mentioned snippets are always there
        ci_snippets: list[Snippet] = pack_items_for_prompt(
            iterable=ci_snippets, string_function=snippet_to_string, token_limit=50_000
        )

        # Cache logic and main fix agent
        fix_issue_result = await fix_issue(
            repo_full_name=repo_full_name,
            query=query,
            access_token=access_token,
            snippets=ci_snippets,
            username=username,
            branch=pull_request.head.ref,
            pulls=pull_request_metadata,
            context=context,
        )
        append_to_ci_history(
            pull_request,
            [
                (
                    FixCIInput(commit_hash=commit.sha, cleaned_logs=cleaned_logs),
                    fix_issue_result,
                )
            ],
        )

        new_history = get_ci_history(pull_request)
        markdown_history = render_history_for_github_comment(pull_request, new_history)

        # Responding to user
        fix_string = (
            f"# Fixes\n > Failing CI in {create_hyperlink(commit)}\n\nIn pull request {pull_request.html_url}\n\n"
        )
        response = (
            initial_message
            + HR
            + format_results(
                fix_issue_result=fix_issue_result,
                fix_string=fix_string,
                repo=repo,
                tracking_id=tracking_id,
                is_comment=True,
            )
        )
        comment.edit(body=response + get_suffix_for_sweep_comment(tracking_id))

        response_with_revert_message = generate_revert_comment_and_format_results(
            repo, fix_issue_result, fix_string, tracking_id
        )
        response_with_revert_message = initial_message + HR + response_with_revert_message
        comment.edit(body=response_with_revert_message + get_suffix_for_sweep_comment(tracking_id))

        return fix_issue_result


async def cleanup_debug_statements(pull_request: PullRequest, tracking_id: str | None = None, force: bool = False):
    if pull_request.state != "open" and not force:
        return
    with log_to_file(tracking_id), IssueErrorHandler(
        pull_request.as_issue(),
        error_title="Error while fixing CI",
        tracking_id=tracking_id,
        do_delete=True,
    ) as monitor:
        logger.info(f"URL: {pull_request.html_url}")
        repo = pull_request.base.repo
        repo_full_name = repo.full_name
        access_token = repo._requester.auth.token
        username = get_actor_from_pull_request(pull_request)
        branch = pull_request.head.ref
        commit: Commit = pull_request.get_commits().reversed[0]

        # TODO: Check if there are any debug statements to remove

        cloned_repo = ClonedRepo.from_repo_full_name(repo_full_name=repo_full_name, branch=branch)
        cloned_repo.safely_checkout_to_non_default_branch(branch_commit_sha=pull_request.head.sha)
        snippets = get_existing_snippets(pull_request, cloned_repo)

        initial_message = f"There are no more CI failures! I'm currently removing debug logs from attempting to fix all the CI failures."
        comment = pull_request.create_issue_comment(initial_message)
        monitor.comment = comment

        if not wildcard_match(username, FIX_CI_ALLOWLIST):
            comment.edit(
                body=f"Hey @{username}, CI auto-fix is currently not enabled for you. Please reach out to the [Sweep team](william@sweep.dev) to enable it."
                + get_suffix_for_sweep_comment(tracking_id)
            )
            return

        pull_request_metadata = generate_pull_request_metadata(pull_request)
        history_str = render_history(pull_request)
        fix_issue_result = await fix_issue(
            repo_full_name=repo_full_name,
            query="Remove debug statements add in this PR.",
            access_token=access_token,
            snippets=snippets,
            username=username,
            branch=pull_request.head.ref,
            pulls=pull_request_metadata,
            context=ChatAgentContext(
                use_case="ci",
                history=history_str,
            ),
        )

        fix_string = f"# Fixes\n > Remove debug statements in {create_hyperlink(commit)}\n\nIn pull request {pull_request.html_url}\n\n"
        response = (
            initial_message
            + HR
            + format_results(
                fix_issue_result=fix_issue_result,
                fix_string=fix_string,
                repo=repo,
                tracking_id=tracking_id,
                is_comment=True,
            )
        )
        comment.edit(body=response + get_suffix_for_sweep_comment(tracking_id))

        response_with_revert_message = generate_revert_comment_and_format_results(
            repo, fix_issue_result, fix_string, tracking_id
        )
        response_with_revert_message = initial_message + HR + response_with_revert_message
        comment.edit(body=response_with_revert_message + get_suffix_for_sweep_comment(tracking_id))

        return fix_issue_result


async def fix_ci_failures(
    pull_request: PullRequest,
    max_attempts: int = 10,
    tracking_id: str | None = None,
    force: bool = False,
):
    # TODO: add better stopping criteria (i.e. making no progress)
    results_list = []
    for i in range(max_attempts):
        fix_issue_result = await fix_ci_failures_once(pull_request, tracking_id=f"{tracking_id}-ci-{i}", force=force)
        if fix_issue_result:
            results_list.append(fix_issue_result)  # Fix attempt
        else:
            logger.info("No CI errors remaining, stopping")
            break  # Nothing else broken
        repo = get_github_repo(repo_full_name=pull_request.base.repo.full_name)
        pull_request = repo.get_pull(pull_request.number)
    # We should only run this if there's not more errors
    # if results_list:
    #     await cleanup_debug_statements(pull_request, tracking_id=f"{tracking_id}-ci-cleanup", force=force)
    return results_list



"""
This is currently exclusively used for the embeddings model.
"""

import os
import random
import ssl

import backoff
import httpx
from anthropic import Anthropic
from loguru import logger
from openai import (
    APITimeoutError,
    AzureOpenAI,
    InternalServerError,
    OpenAI,
    RateLimitError,
)
from openai.types.chat.chat_completion import ChatCompletion

from sweepai.config.server import (
    AZURE_API_KEY,
    AZURE_OPENAI_DEPLOYMENT,
    CURL_CA_BUNDLE,
    DEFAULT_GPT4_MODEL,
    MULTI_REGION_CONFIG,
    OPENAI_API_BASE,
    OPENAI_API_KEY,
    OPENAI_API_TYPE,
    OPENAI_API_VERSION,
    OPENAI_EMBEDDINGS_API_TYPE,
    OPENAI_EMBEDDINGS_AZURE_API_VERSION,
    OPENAI_EMBEDDINGS_AZURE_ENDPOINT,
)
from sweepai.core.entities import Message
from sweepai.utils.cache import file_cache
from sweepai.utils.timer import Timer

OPENAI_TIMEOUT = 120

OPENAI_EXCLUSIVE_MODELS = [
    "gpt-3.5-turbo-1106",
]
SEED = 100

RATE_LIMITS = {
    "australiaeast": 80000,
    "brazilsouth": 0,
    "canadaeast": 80000,
    "eastus": 80000,
    "eastus2": 80000,
    "francecentral": 80000,
    "japaneast": 0,
    "northcentralus": 80000,
    "norwayeast": 150000,
    "southafricanorth": 0,
    "southcentralus": 80000,
    "southindia": 150000,
    "swedencentral": 150000,
    "switzerlandnorth": 0,
    "uksouth": 80000,
    "westeurope": 0,
    "westus": 80000,
}


class OpenAIProxy:
    @file_cache(ignore_params=[])
    def call_openai(
        self,
        model: str,
        messages: list[Message],
        tools: list[str] = [],
        max_tokens: int = 4096,
        temperature: float = 0.0,
        stop_sequences: list[str] = [],
        seed: int = 0,
    ):
        try:
            engine = self.determine_openai_engine(model)
            if OPENAI_API_TYPE is None or engine is None:
                with Timer():
                    response = self.set_openai_default_api_parameters(
                        model=model,
                        messages=messages,
                        max_tokens=max_tokens,
                        temperature=temperature,
                        tools=tools,
                    )
                    return response
            # validity checks for MULTI_REGION_CONFIG
            if (
                MULTI_REGION_CONFIG is None
                or not isinstance(MULTI_REGION_CONFIG, list)
                or len(MULTI_REGION_CONFIG) == 0
                or not isinstance(MULTI_REGION_CONFIG[0], list)
            ):
                if OPENAI_API_TYPE == "azure" or not OPENAI_API_KEY:
                    logger.info(f"Calling {model} with engine {engine} on Azure url {OPENAI_API_BASE}.")
                    try:
                        with Timer():
                            response = self.call_azure_api(
                                model=model,
                                messages=messages,
                                tools=tools,
                                max_tokens=max_tokens,
                                temperature=temperature,
                            )
                            return response.choices[0].message.content
                    except RateLimitError as e:
                        logger.exception(f"Rate Limit Error calling Azure: {e}")
                else:
                    logger.info(f"Calling OpenAI with model {model}.")
                    with Timer():
                        return self.set_openai_default_api_parameters(
                            model=model,
                            messages=messages,
                            max_tokens=max_tokens,
                            temperature=temperature,
                            tools=tools,
                        )
            # multi region config is a list of tuples of (region_url, api_key)
            # we will try each region in order until we get a response
            # randomize the order of the list
            SHUFFLED_MULTI_REGION_CONFIG = random.sample(MULTI_REGION_CONFIG, len(MULTI_REGION_CONFIG))
            for region_url, api_key in SHUFFLED_MULTI_REGION_CONFIG:
                try:
                    logger.info(f"Calling {model} with engine {engine} on Azure url {region_url}.")
                    with Timer():
                        response = self.create_openai_chat_completion(
                            engine=engine,
                            base_url=region_url,
                            api_key=api_key,
                            model=model,
                            messages=messages,
                            tools=tools,
                            max_tokens=max_tokens,
                            temperature=temperature,
                        )
                        return response.choices[0].message.content
                except (RateLimitError, APITimeoutError, InternalServerError) as e:
                    logger.exception(f"RateLimitError calling {region_url}: {e}")
            raise Exception("No Azure regions available")
        except (RateLimitError, APITimeoutError, InternalServerError) as e:
            try:
                with Timer():
                    return self.set_openai_default_api_parameters(
                        model=model,
                        messages=messages,
                        max_tokens=max_tokens,
                        temperature=temperature,
                        tools=tools,
                    )
            except Exception as _e:
                logger.error(f"OpenAI API Key found but error: {_e}")
            logger.error(f"OpenAI API Key not found and Azure Error: {e}")
            # Raise exception to report error
            raise e
        except Exception as e:
            raise e
        return None

    def determine_openai_engine(self, model):
        engine = None
        if model in OPENAI_EXCLUSIVE_MODELS and OPENAI_API_TYPE != "azure":
            logger.info(f"Calling OpenAI exclusive model. {model}")
        elif (
            model == "gpt-4" or model == "gpt-4-0613" or model == "gpt-4-1106-preview" or model == "gpt-4-0125-preview"
        ):
            engine = model
        return engine

    def create_openai_chat_completion(self, engine, base_url, api_key, model, messages, tools, max_tokens, temperature):
        client = AzureOpenAI(
            api_key=api_key,
            azure_endpoint=base_url,
            api_version=OPENAI_API_VERSION,
            azure_deployment=AZURE_OPENAI_DEPLOYMENT,
        )
        if len(tools) == 0:
            response = client.chat.completions.create(
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                timeout=OPENAI_TIMEOUT,
            )
        else:
            response = client.chat.completions.create(
                model=model,
                messages=messages,
                tools=tools,
                max_tokens=max_tokens,
                temperature=temperature,
                timeout=OPENAI_TIMEOUT,
            )
        return response

    def call_azure_api(self, model, messages, tools, max_tokens, temperature) -> ChatCompletion:
        client = AzureOpenAI(
            api_key=AZURE_API_KEY,
            azure_endpoint=OPENAI_API_BASE,
            api_version=OPENAI_API_VERSION,
            azure_deployment=AZURE_OPENAI_DEPLOYMENT,
        )
        if len(tools) == 0:
            response = client.chat.completions.create(
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                timeout=OPENAI_TIMEOUT,
            )
        else:
            response = client.chat.completions.create(
                model=model,
                messages=messages,
                tools=tools,
                max_tokens=max_tokens,
                temperature=temperature,
                timeout=OPENAI_TIMEOUT,
            )
        return response

    @backoff.on_exception(
        backoff.expo,
        exception=(RateLimitError, APITimeoutError, InternalServerError),
        max_tries=3,
        jitter=backoff.random_jitter,
        on_backoff=lambda details: logger.error(
            f"Rate Limit or Timeout Error: {details['tries']} tries. Waiting {details['wait']:.2f} seconds."
        ),
        base=10,
        factor=2,
        max_value=40,
    )
    def set_openai_default_api_parameters(self, model, messages, max_tokens, temperature, tools=[], stop_sequences=[]):
        client = OpenAI(api_key=OPENAI_API_KEY)
        if len(tools) == 0:
            response = client.chat.completions.create(
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                timeout=OPENAI_TIMEOUT,
                seed=SEED,
                stream=True,
            )
            text = ""
            for chunk in response:  # pylint: disable=E1133
                new_content = chunk.choices[0].delta.content
                text += new_content if new_content else ""
                if new_content:
                    print(new_content, end="", flush=True)
            print()  # clear the line
            return text
        else:
            response = client.chat.completions.create(
                model=model,
                messages=messages,
                tools=tools,
                max_tokens=max_tokens,
                temperature=temperature,
                timeout=OPENAI_TIMEOUT,
                seed=SEED,
            )
            return response.choices[0].message.content


def get_client():
    OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")
    OPENAI_API_TYPE = os.environ.get("OPENAI_API_TYPE", "openai")
    OPENAI_API_BASE = os.environ.get("OPENAI_API_BASE", None)
    AZURE_API_KEY = os.environ.get("AZURE_API_KEY", None)
    AZURE_OPENAI_DEPLOYMENT = os.environ.get("AZURE_OPENAI_DEPLOYMENT", None)
    OPENAI_API_VERSION = os.environ.get("OPENAI_API_VERSION", None)
    if OPENAI_API_TYPE == "anthropic":
        client = Anthropic()
        model = "claude-3-opus-20240229"
    if OPENAI_API_TYPE == "openai":
        client = OpenAI(api_key=OPENAI_API_KEY, timeout=90) if OPENAI_API_KEY else None
        model = DEFAULT_GPT4_MODEL
    elif OPENAI_API_TYPE == "azure":
        client = AzureOpenAI(
            azure_endpoint=OPENAI_API_BASE,
            api_key=AZURE_API_KEY,
            api_version=OPENAI_API_VERSION,
        )
        model = AZURE_OPENAI_DEPLOYMENT
    else:
        raise ValueError(f"Invalid OPENAI_API_TYPE: {OPENAI_API_TYPE}")
    return model, client


def get_embeddings_client() -> OpenAI | AzureOpenAI:
    client = None
    if OPENAI_EMBEDDINGS_API_TYPE == "openai":
        client = OpenAI(api_key=OPENAI_API_KEY, timeout=90) if OPENAI_API_KEY else None
    elif OPENAI_EMBEDDINGS_API_TYPE == "azure":
        http_client = None
        if CURL_CA_BUNDLE:
            http_client = httpx.Client(
                verify=ssl.create_default_context(cafile=CURL_CA_BUNDLE),
                proxies={"all://": None},
            )
        client = AzureOpenAI(
            azure_endpoint=OPENAI_EMBEDDINGS_AZURE_ENDPOINT,
            api_key=AZURE_API_KEY,
            api_version=OPENAI_EMBEDDINGS_AZURE_API_VERSION,
            http_client=http_client,
        )
    if not client:
        raise ValueError("No Valid API key found for OpenAI or Azure!")
    return client


def test_openai_proxy():
    openai_proxy = OpenAIProxy()
    response = openai_proxy.call_openai(
        "gpt-4-0125-preview",
        [
            {
                "role": "user",
                "content": "Say this is a test",
            }
        ],
        max_tokens=100,
    )
    print((response))


def test_get_client():
    model, client = get_client()
    client.beta.assistants.create(
        model=model,
        name="Test assistant",
        description="test",
        instructions="Say this is a test",
    )


if __name__ == "__main__":
    model, client = get_client()
    response = client.chat.completions.create(
        model=model,
        messages=[
            Message(
                role="user",
                content="Say this is a test",
            ).to_openai()
        ],
        stream=True,
    )
    print("Generating response...", flush=True)
    text = ""
    for chunk in response:  # pylint: disable=E1133
        new_content = chunk.choices[0].delta.content
        text += new_content if new_content else ""
        if new_content:
            print(new_content, end="", flush=True)
    print()



import difflib


def generate_diff(old_code: str, new_code: str, **kwargs):
    if old_code == new_code:
        return ""
    stripped_old_code = old_code.strip("\n") + "\n"
    stripped_new_code = new_code.strip("\n") + "\n"

    # Split the code into lines, preserving the line endings
    old_lines = old_code.splitlines(keepends=True)
    new_lines = new_code.splitlines(keepends=True)

    # Add a newline character at the end if it's missing
    if not old_code.endswith("\n"):
        old_lines.append("\n")
    if not new_code.endswith("\n"):
        new_lines.append("\n")

    default_kwargs = {"n": 5}
    default_kwargs.update(kwargs)

    diff = difflib.unified_diff(
        stripped_old_code.splitlines(keepends=True), stripped_new_code.splitlines(keepends=True), **kwargs
    )

    diff_result = ""

    for line in diff:
        if not line.endswith("\n"):
            line += "\n"
        diff_result += line

    return diff_result


def generate_diff_statistics(original_code: str, new_code: str) -> str:
    diff = generate_diff(original_code, new_code)
    return diff.count("\n+"), diff.count("\n-")


def remove_whitespace_changes(patch: str) -> str:
    lines = patch.split("\n")
    result = []
    in_hunk = False
    hunk_lines = []
    hunk_header = None

    for line in lines:
        if line.startswith("@@"):
            if in_hunk:
                processed_hunk = process_hunk(hunk_header, hunk_lines)
                if processed_hunk:
                    result.extend(processed_hunk)
            hunk_header = line
            hunk_lines = []
            in_hunk = True
        elif in_hunk:
            hunk_lines.append(line)
        else:
            result.append(line)

    if in_hunk:
        processed_hunk = process_hunk(hunk_header, hunk_lines)
        if processed_hunk:
            result.extend(processed_hunk)

    return "\n".join(filter(None, result))


def should_skip_line(current_line: str, next_line: str) -> bool:
    """This assumes that diffs are ordered in a way that +/- lines are next to each other and - comes before +."""
    if not next_line.startswith("+"):
        return False
    # check if current_line is equal to next_line if both right sides are stripped (don't lstrip because python indentation matters)
    return current_line[1:].rstrip() == next_line[1:].rstrip()


def process_hunk(header: str, hunk: list[str]) -> list[str]:
    processed = []
    changes: list[str] = []
    context_buffer = []

    def flush_changes():
        nonlocal processed, changes, context_buffer
        if changes:
            # Process changes to remove pairs of whitespace-only +/- lines
            cleaned_changes = []
            i = 0
            while i < len(changes):
                line = changes[i]
                if line.startswith("-"):
                    # check if next line is a pure whitespace diff
                    if i + 1 < len(changes):
                        next_line = changes[i + 1]
                        if should_skip_line(current_line=line, next_line=next_line):
                            # Skip both lines
                            i += 2
                            continue
                cleaned_changes.append(line)
                i += 1
            changes = cleaned_changes

            if changes:
                # Check if all remaining changes are whitespace-only
                if all(line[1:].strip() == "" for line in changes):
                    changes = []  # Discard all whitespace-only changes
                else:
                    processed.extend(context_buffer)
                    processed.extend(changes)
            context_buffer = []
            changes = []

    for line in hunk:
        if line.startswith("+") or line.startswith("-"):
            changes.append(line)
        else:
            if changes:
                flush_changes()
            context_buffer.append(line)
            if len(context_buffer) > 3:
                processed.append(context_buffer.pop(0))

    flush_changes()

    if processed:
        return [header] + processed
    else:
        return []



from datetime import datetime
import subprocess

from loguru import logger
import requests
from sweepai.config.server import DEV, GITHUB_BASE_URL

def get_github_time():
    """Fetch current time from GitHub API"""
    response = requests.get(f"https://{GITHUB_BASE_URL}")
    response.raise_for_status()
    # Parse the date from response headers
    github_time = response.headers['date']
    return datetime.strptime(github_time, '%a, %d %b %Y %H:%M:%S GMT')

def set_system_time(dt: datetime):
    return subprocess.run(f'date -s "{dt.strftime("%Y-%m-%d %H:%M:%S")}"', shell=True, check=True)

def sync_time():
    """Main function to sync system time with GitHub"""
    if DEV:
        return
    current_time = datetime.utcnow()
    github_time = get_github_time()
    logger.info(f"Current time: {current_time}, GitHub time: {github_time}, time diff: {(current_time - github_time).total_seconds()} seconds")
    set_system_time(github_time)

    # Verify that the system time is within 5 seconds of the GitHub time
    current_time = datetime.utcnow()
    new_github_time = get_github_time()
    time_diff = (current_time - new_github_time).total_seconds()
    if abs(time_diff) > 5:
        raise Exception(f"System time differs from GitHub time by {time_diff} seconds")
    else:
        logger.info(f"System time synchronized with GitHub: {github_time}, time diff: {time_diff} seconds")

if __name__ == "__main__":
    sync_time() # needs --cap-add SYS_TIME -v /etc/localtime:/etc/localtime



import backoff
from loguru import logger

from sweepai.config.server import COHERE_API_KEY, VOYAGE_API_KEY
from sweepai.core.entities import Message, Snippet
from sweepai.core.llm.chat import LATEST_CLAUDE_SONNET_MODEL, Thread, continuous_llm_calls
from sweepai.dataclasses.search_dataclasses import (
    BaseRerankerResponse,
    BaseRerankerResult,
    Document,
)
from sweepai.utils.cache import file_cache
from sweepai.utils.str_utils import extract_objects_from_string
from sweepai.utils.timer import Timer


@backoff.on_exception(
    backoff.expo,
    Exception,
    max_tries=3,
    jitter=backoff.random_jitter,
)
@file_cache()
def cohere_rerank_call(
    query: str,
    documents: list[str],
    model="rerank-english-v3.0",
    **kwargs,
):
    from cohere import Client

    # Cohere API call with caching
    co = Client(COHERE_API_KEY)
    try:
        return co.rerank(model=model, query=query, documents=documents, **kwargs)
    except Exception as e:
        logger.error(f"Cohere rerank failed: {e}")
        raise e


@file_cache()
def voyage_rerank_call(query: str, documents: list[str], model="rerank-1", **kwargs):
    from voyageai import Client

    vo = Client(api_key=VOYAGE_API_KEY)
    return vo.rerank(query, documents, model=model, **kwargs)


reranking_system_prompt = """You are a powerful code search engine. You must order the list of code blocks from the most relevant to the least relevant to the user's query."""

user_prompt = """
# Code blocks to rank:
{documents}
{directory_summary_string}
# User query:
<query>
{query}
</query>

We are currently ranking code blocks of type {code_block_type}.
- source: This is the source code of the software.
- tests: These are the test files in the repository.
- tools: These are tool configurations and scripts used to build, deploy, or manage the software.
- dependencies: These are dependency configurations. They specify the external libraries and packages required by the software.
- docs: These are documentation files. They provide information and instructions about the software.

# Guidelines:
1. Categorize the relevance of each code block as relevant or irrelevant. The relevance score should be based on how well the code block answers the user query.

2. Ranking Definition:
- "relevant": The code block is relevant to the user query.
- "irrelevant": The code block is completely unrelated to the user query and should be discarded.

3. There may not always be a code block that is relevant to the user query. In this case mark the rest as irrelevant.

4. For ranking, you may look at files given explicitly in the list and sometimes files not given but are relevant to the query from the context of the given code blocks.
    a. For the file paths explicitly mentioned, the <source> must perfectly reference the provided <source> tags within <code_block>.
    b. If you find highly relevant references to some file paths not in the list of code blocks given, the <source> tag should have the path to the file (you can figure this out from the context. If you are unsure, just ignore the file.). Downstream logic will take care of disambiguating such files.

You must now rank the code blocks based on relevance to the user query.

# Response Instructions:

<analysis>
1. Read each file to identify how it relates to the user query. Do not skip any files.
</analysis>

2. Provide the search results in order of relevance, with the most relevant result at the beginning.
<search_result>
<source>[THE FILE PATH OF THE CODE BLOCK]</source>
<relevance_score>relevant or irrelevant</relevance_score>
</search_result>
... # RANK ALL OF THE CODE BLOCKS"""

document_format = """<code_block index={index}>
<source>{file_path}</source>
<code_block_content>
{document_contents}
</code_block_content>
</code_block>"""

def convert_categorical_scores_to_float(scores: list[str]) -> dict[float]:
    # interpolate the scores to a float between 0 and 1
    # relevant = from 0.7 to 1
    # low = from 0.1 to 0.2
    # irrelevant = from 0 to 0.1
    number_of_highs = scores.count("relevant")
    number_of_lows = scores.count("low")
    number_of_irrelevants = scores.count("irrelevant")
    last_high_index = scores.index("relevant") if "relevant" in scores else 0
    last_low_index = scores.index("low") if "low" in scores else 0
    last_irrelevant_index = scores.index("irrelevant") if "irrelevant" in scores else 0
    float_scores = []
    for idx, score in enumerate(scores):
        if score == "relevant":
            offset = (idx - last_high_index) / number_of_highs
            interpolated_high_score = 1.0 - (0.3 * offset)
            interpolated_high_score = round(interpolated_high_score, 2)
            float_scores.append(interpolated_high_score)
        elif score == "low":
            offset = (idx - last_low_index) / number_of_lows
            interpolated_low_score = 0.2 - (0.1 * offset)
            interpolated_low_score = round(interpolated_low_score, 2)
            float_scores.append(interpolated_low_score)
        elif score == "irrelevant":
            offset = (idx - last_irrelevant_index) / number_of_irrelevants
            interpolated_irrelevant_score = 0.1 - (0.1 * offset)
            interpolated_irrelevant_score = round(interpolated_irrelevant_score, 2)
            float_scores.append(interpolated_irrelevant_score)
    return float_scores


def llm_rerank_call(
    query: str,
    documents: list[Document],
    most_relevant_snippet_per_file_path: dict[str, Snippet],
    **kwargs,
):
    search_results = []
    formatted_documents = "\n".join(
        [
            document_format.format(
                index=index,
                document_contents=document.content,
                file_path=document.snippet_denotation,
            )
            for index, document in enumerate(documents)
        ]
    )
    directory_summary_string = ""
    if kwargs.get("all_directory_summaries", dict()):
        all_directory_summaries = kwargs["all_directory_summaries"]
        sorted_directory_summaries = sorted(list(all_directory_summaries.items()), key=lambda x: x[0])
        directory_summary_string = "Here is a summary of the directories where the below code blocks were found:\n\n"
        directory_summary_string += "\n\n".join(
            [
                f"Here is a summary of the {subdir_name} subdirectory:\n{directory_summary}"
                for subdir_name, directory_summary in sorted_directory_summaries
            ]
        )
    code_block_type: str = kwargs.get("code_block_type")
    assert code_block_type, "Code block type must be provided"
    formatted_user_prompt = user_prompt.format(
        query=query,
        documents=formatted_documents,
        directory_summary_string=directory_summary_string,
        code_block_type=code_block_type,
    )
    formatted_user_prompt = formatted_user_prompt.strip()
    messages = [
        Message(
            role="system",
            content=reranking_system_prompt.format(code_block_type=code_block_type),
        )
    ]
    thread = Thread(messages=messages)
    with Timer():
        try:
            # for same query, 10k context:
            # gpt-4o-mini takes ~10s
            # haiku takes 4.5s
            search_results = continuous_llm_calls(
                username=kwargs.get("username", "test_user"),
                content=formatted_user_prompt,
                thread=thread,
                stop_sequences=[
                    "irrelevant</relevance_score>",
                    "low</relevance_score>",
                ],
                MAX_CALLS=1,
                verbose=False,
                model=LATEST_CLAUDE_SONNET_MODEL,
                # can experiment with Claude 3.5 Sonnet later
            )
            if search_results.endswith("</relevance_score>"):
                search_results += "\n</search_result>"
        except Exception as e:
            logger.error(f"LLM rerank failed with exception {e} for query {query}")
            search_results = ""
    reranker_results = parse_search_results_from_string(
        search_results=search_results,
        code_block_type=code_block_type,
        documents=documents,
        most_relevant_snippet_per_file_path=most_relevant_snippet_per_file_path,
    )
    # Ignored for now
    # logger.debug(f"Reranking {code_block_type} for query {query} with results {reranker_results}")
    return reranker_results


def parse_search_results_from_string(
    search_results: str,
    code_block_type: str,
    documents: list[Document],
    most_relevant_snippet_per_file_path: dict[str, Snippet],
) -> BaseRerankerResponse:
    search_objects, failed_extraction = extract_objects_from_string(
        search_results,
        "search_result",
        ["source", "relevance_score"],
        include_closing_tag=False,
    )
    reranker_results = BaseRerankerResponse(results=[])
    if not failed_extraction:
        relevance_scores = []
        for llm_search_response_object in search_objects:
            relevance_score = llm_search_response_object["relevance_score"]
            relevance_scores.append(relevance_score)
        float_relevance_scores = convert_categorical_scores_to_float(relevance_scores)
        for llm_search_response_object, float_relevance_score in zip(search_objects, float_relevance_scores):
            try:
                # clean this up to use most relevant snippet per file path
                for index, document in enumerate(documents):
                    if document.snippet_denotation == llm_search_response_object["source"]:
                        reranker_results.results.append(
                            BaseRerankerResult(
                                snippet_denotation=documents[index].snippet_denotation,
                                relevance_score=float_relevance_score,
                            )
                        )
                        break
                else:
                    if llm_search_response_object["source"] in most_relevant_snippet_per_file_path:
                        reranker_results.results.append(
                            BaseRerankerResult(
                                snippet_denotation=most_relevant_snippet_per_file_path[
                                    llm_search_response_object["source"]
                                ].denotation,
                                relevance_score=float_relevance_score,
                            )
                        )
            except Exception as e:
                logger.error(
                    f"Failed to parse code block {llm_search_response_object['source']} with exception {e}, trying just the file path"
                )
                original_index = next(
                    (
                        index
                        for index, document in enumerate(documents)
                        if document.snippet_denotation.startswith(llm_search_response_object["source"])
                    ),
                    None,
                )
                if original_index is None:
                    logger.error(f"Failed to find code block {llm_search_response_object['source']} in documents")
                    continue
                reranker_results.results.append(
                    BaseRerankerResult(
                        snippet_denotation=documents[original_index].snippet_denotation,
                        relevance_score=float_relevance_score,
                    )
                )
    # add test case for this
    if (code_block_type == "source" or code_block_type == "tests") and len(reranker_results.results) == 0:
        reranker_results.results.append(
            BaseRerankerResult(snippet_denotation=documents[0].snippet_denotation, relevance_score=1)
        )
    return reranker_results


if __name__ == "__main__":
    query = "When is Apple's conference call scheduled?"
    documents = [
        Document(
            snippet_denotation="The Mediterranean Diet",
            content="The Mediterranean diet emphasizes fish, olive oil, and vegetables, believed to reduce chronic diseases.",
        ),
        Document(
            snippet_denotation="Photosynthesis in Plants",
            content="Photosynthesis in plants converts light energy into glucose and produces essential oxygen.",
        ),
        Document(
            snippet_denotation="20th-century Innovations",
            content="20th-century innovations, from radios to smartphones, centered on electronic advancements.",
        ),
        Document(
            snippet_denotation="Rivers",
            content="Rivers provide water, irrigation, and habitat for aquatic species, vital for ecosystems.",
        ),
        Document(
            snippet_denotation="Latest News",
            content="Apple’s conference call to discuss fourth fiscal quarter results and business updates is scheduled for Thursday, November 2, 2023 at 2:00 p.m. PT / 5:00 p.m. ET.",
        ),
        Document(
            snippet_denotation="Shakespeare's Works",
            content="Shakespeare's works, like 'Hamlet' and 'A Midsummer Night's Dream,' endure in literature.",
        ),
    ]

    reranking = llm_rerank_call(query, documents)
    for r in reranking.results:
        print(f"Document: {r.source}")
        print(f"Relevance Score: {r.relevance_score}")



import hashlib
import inspect
import os
import pickle
import time

from diskcache import Cache
from loguru import logger

from sweepai.config.server import CACHE_DIRECTORY, FILE_CACHE_DISABLED

TEST_BOT_NAME = "sweep-nightly[bot]"
MAX_DEPTH = 10
# if DEBUG:
#     logger.debug("File cache is disabled.")


def recursive_hash(value, depth=0, ignore_params=[]):
    """Hash primitives recursively with maximum depth."""
    if depth > MAX_DEPTH:
        return hashlib.md5("max_depth_reached".encode()).hexdigest()

    if isinstance(value, (int, float, str, bool, bytes)):
        return hashlib.md5(str(value).encode()).hexdigest()
    elif isinstance(value, (list, tuple)):
        return hashlib.md5(
            "".join([recursive_hash(item, depth + 1, ignore_params) for item in value]).encode()
        ).hexdigest()
    elif isinstance(value, dict):
        return hashlib.md5(
            "".join(
                [
                    recursive_hash(key, depth + 1, ignore_params) + recursive_hash(val, depth + 1, ignore_params)
                    for key, val in value.items()
                    if key not in ignore_params
                ]
            ).encode()
        ).hexdigest()
    elif hasattr(value, "cache_key") and value.cache_key:
        return value.cache_key
    elif hasattr(value, "__dict__") and value.__class__.__name__ not in ignore_params:
        return recursive_hash(value.__dict__, depth + 1, ignore_params)
    else:
        return hashlib.md5("unknown".encode()).hexdigest()


def hash_code(code):
    return hashlib.md5(code.encode()).hexdigest()


def file_cache(ignore_params=[], ignore_contents=False, verbose=False):
    """Decorator to cache function output based on its inputs, ignoring specified parameters.
    Ignore parameters are used to avoid caching on non-deterministic inputs, such as timestamps.
    We can also ignore parameters that are slow to serialize/constant across runs, such as large objects.
    """

    def decorator(func):
        if FILE_CACHE_DISABLED:
            return func
        func_source_code_hash = hash_code(inspect.getsource(func)) if not ignore_contents else ""

        def wrapper(*args, **kwargs):
            if kwargs.get("do_not_use_file_cache", False):
                return func(*args, **kwargs)
            cache_dir = CACHE_DIRECTORY + "/file_cache"
            os.makedirs(cache_dir, exist_ok=True)
            result = None

            # Convert args to a dictionary based on the function's signature
            args_names = func.__code__.co_varnames[: func.__code__.co_argcount]
            args_dict = dict(zip(args_names, args))

            # Remove ignored params
            kwargs_clone = kwargs.copy()
            for param in ignore_params:
                args_dict.pop(param, None)
                kwargs_clone.pop(param, None)

            # Create hash based on function name, input arguments, and function source code
            arg_hash = (
                recursive_hash(args_dict, ignore_params=ignore_params)
                + recursive_hash(kwargs_clone, ignore_params=ignore_params)
                + func_source_code_hash
            )
            cache_key = f"{func.__module__}_{func.__name__}_{arg_hash}"
            cache_file = os.path.join(cache_dir, f"{cache_key}.pickle")
            try:
                # If cache exists, load and return it
                if os.path.exists(cache_file):
                    if verbose:
                        print("Used cache for function: " + func.__name__)
                    with open(cache_file, "rb") as f:
                        result = pickle.load(f)
            except Exception:
                logger.info("Unpickling failed")
            # Otherwise, call the function and save its result to the cache
            if result is None:
                result = func(*args, **kwargs)
            # hydrate both caches in all cases
            if isinstance(result, Exception):
                logger.info(f"Function {func.__name__} returned an exception")
            elif not os.path.exists(cache_file):
                try:
                    with open(cache_file, "wb") as f:
                        pickle.dump(result, f)
                except Exception as e:
                    logger.info(f"Pickling failed: {e}")
            return result

        return wrapper

    return decorator


def create_cache_from_file(file_path: str):
    return Cache(f"{CACHE_DIRECTORY}/{os.path.splitext(os.path.basename(file_path))[0]}")


def create_cache():
    file = os.path.relpath(inspect.currentframe().f_back.f_code.co_filename)
    return create_cache_from_file(file)


if __name__ == "__main__":

    @file_cache()
    def test_func(a, b):
        time.sleep(3)
        return a + b

    print("Running...")
    print(test_func(1, 1))
    print("Running again")
    print(test_func(1, 1))
    print("Done")
